from mlflow_oidc_auth.repository.experiment_permission import ExperimentPermissionRepository
from mlflow_oidc_auth.repository.experiment_permission_group import ExperimentPermissionGroupRepository
from mlflow_oidc_auth.repository.group import GroupRepository
from mlflow_oidc_auth.repository.prompt_permission_group import PromptPermissionGroupRepository
from mlflow_oidc_auth.repository.registered_model_permission import RegisteredModelPermissionRepository
from mlflow_oidc_auth.repository.registered_model_permission_group import RegisteredModelPermissionGroupRepository
from mlflow_oidc_auth.repository.user import UserRepository
from mlflow_oidc_auth.repository.experiment_permission_regex import ExperimentPermissionRegexRepository
from mlflow_oidc_auth.repository.experiment_permission_regex_group import ExperimentPermissionGroupRegexRepository
from mlflow_oidc_auth.repository.registered_model_permission_regex import RegisteredModelPermissionRegexRepository
from mlflow_oidc_auth.repository.registered_model_permission_regex_group import RegisteredModelGroupRegexPermissionRepository

from mlflow_oidc_auth.repository.scorer_permission import ScorerPermissionRepository
from mlflow_oidc_auth.repository.scorer_permission_group import ScorerPermissionGroupRepository
from mlflow_oidc_auth.repository.scorer_permission_regex import ScorerPermissionRegexRepository
from mlflow_oidc_auth.repository.scorer_permission_regex_group import ScorerPermissionGroupRegexRepository

from mlflow_oidc_auth.repository.gateway_secret_permissions import GatewaySecretPermissionRepository
from mlflow_oidc_auth.repository.gateway_secret_regex_permissions import GatewaySecretPermissionRegexRepository
from mlflow_oidc_auth.repository.gateway_secret_group_permissions import GatewaySecretGroupPermissionRepository
from mlflow_oidc_auth.repository.gateway_secret_group_regex_permissions import GatewaySecretPermissionGroupRegexRepository

from mlflow_oidc_auth.repository.gateway_endpoint_permissions import GatewayEndpointPermissionRepository
from mlflow_oidc_auth.repository.gateway_endpoint_regex_permissions import GatewayEndpointPermissionRegexRepository
from mlflow_oidc_auth.repository.gateway_endpoint_group_permissions import GatewayEndpointGroupPermissionRepository
from mlflow_oidc_auth.repository.gateway_endpoint_group_regex_permissions import GatewayEndpointPermissionGroupRegexRepository

from mlflow_oidc_auth.repository.gateway_model_definition_permissions import GatewayModelDefinitionPermissionRepository
from mlflow_oidc_auth.repository.gateway_model_definition_regex_permissions import GatewayModelDefinitionPermissionRegexRepository
from mlflow_oidc_auth.repository.gateway_model_definition_group_permissions import GatewayModelDefinitionGroupPermissionRepository
from mlflow_oidc_auth.repository.gateway_model_definition_group_regex_permissions import GatewayModelDefinitionPermissionGroupRegexRepository

__all__ = [
    "ExperimentPermissionRepository",
    "ExperimentPermissionGroupRepository",
    "GroupRepository",
    "PromptPermissionGroupRepository",
    "RegisteredModelPermissionRepository",
    "RegisteredModelPermissionGroupRepository",
    "UserRepository",
    "ExperimentPermissionRegexRepository",
    "ExperimentPermissionGroupRegexRepository",
    "RegisteredModelPermissionRegexRepository",
    "RegisteredModelGroupRegexPermissionRepository",
    "ScorerPermissionRepository",
    "ScorerPermissionGroupRepository",
    "ScorerPermissionRegexRepository",
    "ScorerPermissionGroupRegexRepository",
    "GatewaySecretPermissionRepository",
    "GatewaySecretPermissionRegexRepository",
    "GatewaySecretGroupPermissionRepository",
    "GatewaySecretPermissionGroupRegexRepository",
    "GatewayEndpointPermissionRepository",
    "GatewayEndpointPermissionRegexRepository",
    "GatewayEndpointGroupPermissionRepository",
    "GatewayEndpointPermissionGroupRegexRepository",
    "GatewayModelDefinitionPermissionRepository",
    "GatewayModelDefinitionPermissionRegexRepository",
    "GatewayModelDefinitionGroupPermissionRepository",
    "GatewayModelDefinitionPermissionGroupRegexRepository",
]



from typing import Any, Dict, List

from fastapi import APIRouter, Depends, HTTPException, Path
from fastapi.responses import JSONResponse

from mlflow_oidc_auth.dependencies import check_gateway_secret_manage_permission
from mlflow_oidc_auth.logger import get_logger
from mlflow_oidc_auth.models.group import GroupPermissionEntry
from mlflow_oidc_auth.models.permission import UserPermission
from mlflow_oidc_auth.routers._prefix import GATEWAY_PERMISSIONS_ROUTER_PREFIX
from mlflow_oidc_auth.store import store
from mlflow_oidc_auth.utils import fetch_all_gateway_secrets, get_is_admin, get_username
from mlflow_oidc_auth.utils.batch_permissions import filter_manageable_gateway_secrets

logger = get_logger()

gateway_secret_permissions_router = APIRouter(
    prefix=f"{GATEWAY_PERMISSIONS_ROUTER_PREFIX}/secrets",
    tags=["gateway secret permissions"],
    responses={
        403: {"description": "Forbidden - Insufficient permissions"},
        404: {"description": "Resource not found"},
    },
)


LIST_SECRETS = ""

GATEWAY_SECRET_USER_PERMISSIONS = "/{name:path}/users"
GATEWAY_SECRET_GROUP_PERMISSIONS = "/{name:path}/groups"


@gateway_secret_permissions_router.get(
    GATEWAY_SECRET_USER_PERMISSIONS,
    response_model=List[UserPermission],
    summary="List users with permissions for a gateway secret",
    description="Retrieves a list of users who have permissions for the specified gateway secret.",
)
async def get_gateway_secret_users(
    name: str = Path(..., description="The gateway secret name to get permissions for"),
    _: None = Depends(check_gateway_secret_manage_permission),
) -> List[UserPermission]:
    list_users = store.list_users(all=True)

    users: List[UserPermission] = []
    for user in list_users:
        user_secrets = {}
        if hasattr(user, "gateway_secret_permissions") and user.gateway_secret_permissions:
            user_secrets = {g.secret_id: g.permission for g in user.gateway_secret_permissions}

        if name in user_secrets:
            users.append(
                UserPermission(
                    name=user.username,
                    permission=user_secrets[name],
                    kind="service-account" if user.is_service_account else "user",
                )
            )
    return users


@gateway_secret_permissions_router.get(
    GATEWAY_SECRET_GROUP_PERMISSIONS,
    response_model=List[GroupPermissionEntry],
    summary="List groups with permissions for a gateway secret",
    description="Retrieves a list of groups who have permissions for the specified gateway secret.",
)
async def get_gateway_secret_groups(
    name: str = Path(..., description="The gateway secret name to get permissions for"),
    _: None = Depends(check_gateway_secret_manage_permission),
) -> List[GroupPermissionEntry]:
    """List all groups with permissions for a specific gateway secret."""
    try:
        groups = store.gateway_secret_group_repo.list_groups_for_secret(name)
        return [GroupPermissionEntry(name=group_name, permission=permission) for group_name, permission in groups]
    except Exception as e:
        logger.error(f"Error retrieving gateway secret group permissions: {e}")
        raise HTTPException(status_code=500, detail="Failed to retrieve gateway secret group permissions")


@gateway_secret_permissions_router.get(
    LIST_SECRETS,
    response_model=List[Dict[str, Any]],
    summary="List all gateway secrets",
    description="Retrieves a list of all gateway secrets from MLflow.",
)
async def list_gateway_secrets(username: str = Depends(get_username), is_admin: bool = Depends(get_is_admin)) -> JSONResponse:
    """
    List gateway secrets accessible to the authenticated user.

    This endpoint returns gateway secrets from MLflow's AI Gateway based on user permissions:
    - Administrators can see all gateway secrets
    - Regular users only see gateway secrets they can manage

    Parameters:
    -----------
    username : str
        The authenticated username (injected by dependency).
    is_admin : bool
        Whether the user has admin privileges (injected by dependency).

    Returns:
    --------
    JSONResponse
        A JSON response containing the list of accessible gateway secrets.

    Raises:
    -------
    HTTPException
        If there is an error retrieving the secrets.
    """
    try:
        # Fetch all gateway secrets from MLflow's tracking store
        all_secrets = fetch_all_gateway_secrets()

        if is_admin:
            # Admins can see all secrets
            secrets = all_secrets
        else:
            # Regular users only see secrets they can manage
            secrets = filter_manageable_gateway_secrets(username, all_secrets)

        # Format the response to match the expected frontend schema
        return JSONResponse(
            content=[
                {
                    "key": secret.get("secret_name") or secret.get("name") or secret.get("key", ""),
                }
                for secret in secrets
            ]
        )
    except Exception as e:
        logger.error(f"Error listing gateway secrets: {e}")
        raise HTTPException(status_code=500, detail="Failed to retrieve gateway secrets")



"""Tests for gateway group-level permission repositories (endpoint, secret, model definition)."""

import pytest
from unittest.mock import MagicMock, patch
from sqlalchemy.exc import IntegrityError, MultipleResultsFound, NoResultFound
from mlflow.exceptions import MlflowException

from mlflow_oidc_auth.repository.gateway_endpoint_group_permissions import (
    GatewayEndpointGroupPermissionRepository,
)
from mlflow_oidc_auth.repository.gateway_secret_group_permissions import (
    GatewaySecretGroupPermissionRepository,
)
from mlflow_oidc_auth.repository.gateway_model_definition_group_permissions import (
    GatewayModelDefinitionGroupPermissionRepository,
)

# ---------------------------------------------------------------------------
# Fixtures
# ---------------------------------------------------------------------------


@pytest.fixture
def session():
    """Create a mock session with context-manager support."""
    s = MagicMock()
    s.__enter__.return_value = s
    s.__exit__.return_value = None
    return s


@pytest.fixture
def session_maker(session):
    """Create a mock session maker."""
    return MagicMock(return_value=session)


# ---------------------------------------------------------------------------
# Parameterised definitions for each resource type
# ---------------------------------------------------------------------------

RESOURCE_CONFIGS = [
    pytest.param(
        GatewayEndpointGroupPermissionRepository,
        "mlflow_oidc_auth.repository.gateway_endpoint_group_permissions",
        "endpoint_id",
        "SqlGatewayEndpointGroupPermission",
        "list_groups_for_endpoint",
        id="endpoint",
    ),
    pytest.param(
        GatewaySecretGroupPermissionRepository,
        "mlflow_oidc_auth.repository.gateway_secret_group_permissions",
        "secret_id",
        "SqlGatewaySecretGroupPermission",
        "list_groups_for_secret",
        id="secret",
    ),
    pytest.param(
        GatewayModelDefinitionGroupPermissionRepository,
        "mlflow_oidc_auth.repository.gateway_model_definition_group_permissions",
        "model_definition_id",
        "SqlGatewayModelDefinitionGroupPermission",
        "list_groups_for_model_definition",
        id="model_definition",
    ),
]


# ---------------------------------------------------------------------------
# Tests — _get_group_permission
# ---------------------------------------------------------------------------


@pytest.mark.parametrize("repo_cls,mod,field,sql_cls,list_groups_method", RESOURCE_CONFIGS)
class TestGetGroupPermission:
    """Tests for _get_group_permission private helper."""

    def test_found(self, session_maker, session, repo_cls, mod, field, sql_cls, list_groups_method):
        """Test successful lookup returns the row."""
        repo = repo_cls(session_maker)
        perm = MagicMock()
        session.query().join().filter().one.return_value = perm
        result = repo._get_group_permission(session, "res-1", "devs")
        assert result == perm

    def test_not_found(self, session_maker, session, repo_cls, mod, field, sql_cls, list_groups_method):
        """Test NoResultFound raises MlflowException."""
        repo = repo_cls(session_maker)
        session.query().join().filter().one.side_effect = NoResultFound()
        with pytest.raises(MlflowException) as exc:
            repo._get_group_permission(session, "res-1", "devs")
        assert exc.value.error_code == "RESOURCE_DOES_NOT_EXIST"

    def test_multiple_found(self, session_maker, session, repo_cls, mod, field, sql_cls, list_groups_method):
        """Test MultipleResultsFound raises MlflowException."""
        repo = repo_cls(session_maker)
        session.query().join().filter().one.side_effect = MultipleResultsFound()
        with pytest.raises(MlflowException) as exc:
            repo._get_group_permission(session, "res-1", "devs")
        assert exc.value.error_code == "INVALID_STATE"


# ---------------------------------------------------------------------------
# Tests — grant_group_permission
# ---------------------------------------------------------------------------


@pytest.mark.parametrize("repo_cls,mod,field,sql_cls,list_groups_method", RESOURCE_CONFIGS)
class TestGrantGroupPermission:
    """Tests for grant_group_permission."""

    def test_success(self, session_maker, session, repo_cls, mod, field, sql_cls, list_groups_method):
        """Test successful grant returns the entity."""
        repo = repo_cls(session_maker)
        group = MagicMock(id=10)
        perm = MagicMock()
        perm.to_mlflow_entity.return_value = "entity"
        with (
            patch(f"{mod}.get_group", return_value=group),
            patch(f"{mod}.{sql_cls}", return_value=perm),
            patch(f"{mod}._validate_permission"),
        ):
            result = repo.grant_group_permission("devs", "res-1", "READ")
        assert result == "entity"
        session.add.assert_called_once_with(perm)
        session.flush.assert_called_once()

    def test_integrity_error(self, session_maker, session, repo_cls, mod, field, sql_cls, list_groups_method):
        """Test IntegrityError raises RESOURCE_ALREADY_EXISTS."""
        repo = repo_cls(session_maker)
        group = MagicMock(id=10)
        session.flush.side_effect = IntegrityError("stmt", "params", "orig")
        with (
            patch(f"{mod}.get_group", return_value=group),
            patch(f"{mod}.{sql_cls}", return_value=MagicMock()),
            patch(f"{mod}._validate_permission"),
        ):
            with pytest.raises(MlflowException) as exc:
                repo.grant_group_permission("devs", "res-1", "READ")
        assert exc.value.error_code == "RESOURCE_ALREADY_EXISTS"


# ---------------------------------------------------------------------------
# Tests — get_group_permission_for_user
# ---------------------------------------------------------------------------


@pytest.mark.parametrize("repo_cls,mod,field,sql_cls,list_groups_method", RESOURCE_CONFIGS)
class TestGetGroupPermissionForUser:
    """Tests for get_group_permission_for_user."""

    def test_success(self, session_maker, session, repo_cls, mod, field, sql_cls, list_groups_method):
        """Test successful get returns the entity."""
        repo = repo_cls(session_maker)
        perm = MagicMock()
        perm.to_mlflow_entity.return_value = "entity"
        with patch.object(repo, "_get_group_permission", return_value=perm):
            assert repo.get_group_permission_for_user("res-1", "devs") == "entity"


# ---------------------------------------------------------------------------
# Tests — list_permissions_for_group
# ---------------------------------------------------------------------------


@pytest.mark.parametrize("repo_cls,mod,field,sql_cls,list_groups_method", RESOURCE_CONFIGS)
class TestListPermissionsForGroup:
    """Tests for list_permissions_for_group."""

    def test_returns_entities(self, session_maker, session, repo_cls, mod, field, sql_cls, list_groups_method):
        """Test list returns mapped entities."""
        repo = repo_cls(session_maker)
        group = MagicMock(id=10)
        perm1 = MagicMock()
        perm1.to_mlflow_entity.return_value = "e1"
        perm2 = MagicMock()
        perm2.to_mlflow_entity.return_value = "e2"
        session.query().filter().all.return_value = [perm1, perm2]
        with patch(f"{mod}.get_group", return_value=group):
            result = repo.list_permissions_for_group("devs")
        assert result == ["e1", "e2"]

    def test_empty(self, session_maker, session, repo_cls, mod, field, sql_cls, list_groups_method):
        """Test list returns empty list when no permissions exist."""
        repo = repo_cls(session_maker)
        group = MagicMock(id=10)
        session.query().filter().all.return_value = []
        with patch(f"{mod}.get_group", return_value=group):
            assert repo.list_permissions_for_group("devs") == []


# ---------------------------------------------------------------------------
# Tests — update_group_permission
# ---------------------------------------------------------------------------


@pytest.mark.parametrize("repo_cls,mod,field,sql_cls,list_groups_method", RESOURCE_CONFIGS)
class TestUpdateGroupPermission:
    """Tests for update_group_permission."""

    def test_success(self, session_maker, session, repo_cls, mod, field, sql_cls, list_groups_method):
        """Test successful update sets permission and flushes."""
        repo = repo_cls(session_maker)
        perm = MagicMock()
        perm.to_mlflow_entity.return_value = "entity"
        with (
            patch.object(repo, "_get_group_permission", return_value=perm),
            patch(f"{mod}._validate_permission"),
        ):
            result = repo.update_group_permission("devs", "res-1", "EDIT")
        assert result == "entity"
        assert perm.permission == "EDIT"
        session.flush.assert_called_once()


# ---------------------------------------------------------------------------
# Tests — revoke_group_permission
# ---------------------------------------------------------------------------


@pytest.mark.parametrize("repo_cls,mod,field,sql_cls,list_groups_method", RESOURCE_CONFIGS)
class TestRevokeGroupPermission:
    """Tests for revoke_group_permission."""

    def test_success(self, session_maker, session, repo_cls, mod, field, sql_cls, list_groups_method):
        """Test successful revoke deletes and flushes."""
        repo = repo_cls(session_maker)
        perm = MagicMock()
        with patch.object(repo, "_get_group_permission", return_value=perm):
            repo.revoke_group_permission("devs", "res-1")
        session.delete.assert_called_once_with(perm)
        session.flush.assert_called_once()


# ---------------------------------------------------------------------------
# Tests — list_groups_for_*  (extra method unique to group repos)
# ---------------------------------------------------------------------------


@pytest.mark.parametrize("repo_cls,mod,field,sql_cls,list_groups_method", RESOURCE_CONFIGS)
class TestListGroupsForResource:
    """Tests for the list_groups_for_<resource> method."""

    def test_returns_tuples(self, session_maker, session, repo_cls, mod, field, sql_cls, list_groups_method):
        """Test that list_groups_for_* returns (group_name, permission) tuples."""
        repo = repo_cls(session_maker)
        perm1 = MagicMock(permission="READ")
        perm2 = MagicMock(permission="MANAGE")
        session.query().join().filter().all.return_value = [
            (perm1, "group-a"),
            (perm2, "group-b"),
        ]
        method = getattr(repo, list_groups_method)
        result = method("res-1")
        assert result == [("group-a", "READ"), ("group-b", "MANAGE")]

    def test_empty(self, session_maker, session, repo_cls, mod, field, sql_cls, list_groups_method):
        """Test empty result set."""
        repo = repo_cls(session_maker)
        session.query().join().filter().all.return_value = []
        method = getattr(repo, list_groups_method)
        assert method("res-1") == []



"""
Tests for the trash router.
"""

from unittest.mock import MagicMock, patch

import pytest
from fastapi.testclient import TestClient
from mlflow.entities import ViewType

from mlflow_oidc_auth.routers.trash import (
    list_deleted_experiments,
    list_deleted_runs,
    permanently_delete_all_trashed_entities,
    restore_experiment,
    restore_run,
)


class TestListDeletedExperimentsEndpoint:
    """Test the list deleted experiments endpoint functionality."""

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash.fetch_all_experiments")
    async def test_list_deleted_experiments_success(self, mock_fetch_all_experiments):
        """Test successfully listing deleted experiments as admin."""
        # Mock deleted experiments
        mock_deleted_experiment = MagicMock()
        mock_deleted_experiment.experiment_id = "123"
        mock_deleted_experiment.name = "Deleted Experiment"
        mock_deleted_experiment.lifecycle_stage = "deleted"
        mock_deleted_experiment.artifact_location = "/tmp/artifacts/123"
        mock_deleted_experiment.tags = {"tag1": "value1"}
        mock_deleted_experiment.creation_time = 1000000
        mock_deleted_experiment.last_update_time = 2000000

        mock_fetch_all_experiments.return_value = [mock_deleted_experiment]

        # Call the function
        result = await list_deleted_experiments(admin_username="admin@example.com")

        # Verify call
        mock_fetch_all_experiments.assert_called_once_with(view_type=ViewType.DELETED_ONLY)

        # Verify response
        assert result.status_code == 200
        # Access the JSON content from the JSONResponse
        import json

        response_data = json.loads(result.body)
        assert "deleted_experiments" in response_data
        assert len(response_data["deleted_experiments"]) == 1
        assert response_data["deleted_experiments"][0]["experiment_id"] == "123"
        assert response_data["deleted_experiments"][0]["name"] == "Deleted Experiment"

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash.fetch_all_experiments")
    async def test_list_deleted_experiments_empty(self, mock_fetch_all_experiments):
        """Test listing deleted experiments when none exist."""
        mock_fetch_all_experiments.return_value = []

        # Call the function
        result = await list_deleted_experiments(admin_username="admin@example.com")

        # Verify call
        mock_fetch_all_experiments.assert_called_once_with(view_type=ViewType.DELETED_ONLY)

        # Verify response
        assert result.status_code == 200
        import json

        response_data = json.loads(result.body)
        assert "deleted_experiments" in response_data
        assert len(response_data["deleted_experiments"]) == 0

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash.fetch_all_experiments")
    async def test_list_deleted_experiments_error(self, mock_fetch_all_experiments):
        """Test error handling when fetching deleted experiments fails."""
        mock_fetch_all_experiments.side_effect = Exception("MLflow error")

        # Call the function and verify it raises HTTPException
        from fastapi import HTTPException

        with pytest.raises(HTTPException) as excinfo:
            await list_deleted_experiments(admin_username="admin@example.com")

        assert excinfo.value.status_code == 500
        assert excinfo.value.detail == "Failed to retrieve deleted experiments"

    def test_list_deleted_experiments_integration_admin(self, admin_client: TestClient):
        """Test the endpoint through FastAPI test client as admin."""
        # Mock the fetch function
        with patch("mlflow_oidc_auth.routers.trash.fetch_all_experiments") as mock_fetch:
            mock_experiment = MagicMock()
            mock_experiment.experiment_id = "123"
            mock_experiment.name = "Deleted Experiment"
            mock_experiment.lifecycle_stage = "deleted"
            mock_experiment.artifact_location = "/tmp/artifacts/123"
            mock_experiment.tags = {"tag1": "value1"}
            mock_experiment.creation_time = 1000000
            mock_experiment.last_update_time = 2000000
            mock_fetch.return_value = [mock_experiment]

            response = admin_client.get("/oidc/trash/experiments")

            assert response.status_code == 200
            data = response.json()
            assert "deleted_experiments" in data
            assert len(data["deleted_experiments"]) == 1
            assert data["deleted_experiments"][0]["experiment_id"] == "123"
            assert data["deleted_experiments"][0]["name"] == "Deleted Experiment"
            assert data["deleted_experiments"][0]["lifecycle_stage"] == "deleted"

    def test_list_deleted_experiments_integration_non_admin(self, client: TestClient):
        """Test the endpoint through FastAPI test client as non-admin (should be forbidden)."""
        response = client.get("/oidc/trash/experiments")

        # Should be forbidden for non-admin users
        assert response.status_code == 403


class TestListDeletedRunsEndpoint:
    """Tests for listing deleted runs."""

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_list_deleted_runs_success(self, mock_get_store):
        backend_store = MagicMock()
        backend_store._get_deleted_runs.return_value = ["run-1", "run-2"]

        run_deleted = MagicMock()
        run_deleted.info.run_id = "run-1"
        run_deleted.info.experiment_id = "exp-1"
        run_deleted.info.run_name = "name-1"
        run_deleted.info.status = "FINISHED"
        run_deleted.info.start_time = 1
        run_deleted.info.end_time = 2
        run_deleted.info.lifecycle_stage = "deleted"

        run_active = MagicMock()
        run_active.info.run_id = "run-2"
        run_active.info.experiment_id = "exp-2"
        run_active.info.run_name = "name-2"
        run_active.info.status = "FINISHED"
        run_active.info.start_time = 3
        run_active.info.end_time = 4
        run_active.info.lifecycle_stage = "active"

        backend_store.get_run.side_effect = [run_deleted, run_active]
        mock_get_store.return_value = backend_store

        result = await list_deleted_runs(admin_username="admin@example.com", experiment_ids=None, older_than=None)

        backend_store._get_deleted_runs.assert_called_once()
        assert result.status_code == 200
        import json

        payload = json.loads(result.body)
        assert payload["deleted_runs"] == [
            {
                "run_id": "run-1",
                "experiment_id": "exp-1",
                "run_name": "name-1",
                "status": "FINISHED",
                "start_time": 1,
                "end_time": 2,
                "lifecycle_stage": "deleted",
            }
        ]

    @pytest.mark.asyncio
    async def test_list_deleted_runs_invalid_older_than(self):
        result = await list_deleted_runs(admin_username="admin@example.com", experiment_ids=None, older_than="bad")
        assert result.status_code == 400

    def test_list_deleted_runs_integration_admin(self, admin_client: TestClient):
        with patch("mlflow_oidc_auth.routers.trash._get_store") as mock_get_store:
            backend_store = MagicMock()
            backend_store._get_deleted_runs.return_value = ["run-1"]

            run_deleted = MagicMock()
            run_deleted.info.run_id = "run-1"
            run_deleted.info.experiment_id = "exp-1"
            run_deleted.info.run_name = "deleted-run"
            run_deleted.info.status = "FINISHED"
            run_deleted.info.start_time = 10
            run_deleted.info.end_time = 20
            run_deleted.info.lifecycle_stage = "deleted"

            backend_store.get_run.return_value = run_deleted
            mock_get_store.return_value = backend_store

            response = admin_client.get("/oidc/trash/runs")
            assert response.status_code == 200
            assert response.json()["deleted_runs"][0]["run_id"] == "run-1"

    def test_list_deleted_runs_integration_non_admin(self, client: TestClient):
        response = client.get("/oidc/trash/runs")
        assert response.status_code == 403


class TestRestoreExperimentEndpoint:
    """Tests for restoring experiments."""

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_restore_experiment_success(self, mock_get_store):
        backend_store = MagicMock()
        deleted = MagicMock()
        deleted.lifecycle_stage = "deleted"
        deleted.experiment_id = "123"
        deleted.name = "exp"
        deleted.last_update_time = 1

        restored = MagicMock()
        restored.lifecycle_stage = "active"
        restored.experiment_id = "123"
        restored.name = "exp"
        restored.last_update_time = 2

        backend_store.get_experiment.side_effect = [deleted, restored]
        mock_get_store.return_value = backend_store

        result = await restore_experiment(experiment_id="123", admin_username="admin@example.com")
        backend_store.restore_experiment.assert_called_once_with("123")
        assert result.status_code == 200

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_restore_experiment_not_deleted(self, mock_get_store):
        backend_store = MagicMock()
        active = MagicMock()
        active.lifecycle_stage = "active"
        backend_store.get_experiment.return_value = active
        mock_get_store.return_value = backend_store

        result = await restore_experiment(experiment_id="123", admin_username="admin@example.com")
        assert result.status_code == 400

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_restore_experiment_not_found(self, mock_get_store):
        backend_store = MagicMock()
        backend_store.get_experiment.side_effect = Exception("not found")
        mock_get_store.return_value = backend_store

        result = await restore_experiment(experiment_id="missing", admin_username="admin@example.com")
        assert result.status_code == 404


class TestRestoreRunEndpoint:
    """Tests for restoring runs."""

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_restore_run_success(self, mock_get_store):
        backend_store = MagicMock()
        deleted = MagicMock()
        deleted.info.lifecycle_stage = "deleted"
        deleted.info.run_id = "run-1"
        deleted.info.experiment_id = "exp-1"
        deleted.info.run_name = "r"
        deleted.info.status = "FINISHED"

        restored = MagicMock()
        restored.info.lifecycle_stage = "active"
        restored.info.run_id = "run-1"
        restored.info.experiment_id = "exp-1"
        restored.info.run_name = "r"
        restored.info.status = "FINISHED"

        backend_store.get_run.side_effect = [deleted, restored]
        mock_get_store.return_value = backend_store

        result = await restore_run(run_id="run-1", admin_username="admin@example.com")
        backend_store.restore_run.assert_called_once_with("run-1")
        assert result.status_code == 200

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_restore_run_not_deleted(self, mock_get_store):
        backend_store = MagicMock()
        active = MagicMock()
        active.info.lifecycle_stage = "active"
        backend_store.get_run.return_value = active
        mock_get_store.return_value = backend_store

        result = await restore_run(run_id="run-1", admin_username="admin@example.com")
        assert result.status_code == 400

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_restore_run_not_found(self, mock_get_store):
        backend_store = MagicMock()
        backend_store.get_run.side_effect = Exception("missing")
        mock_get_store.return_value = backend_store

        result = await restore_run(run_id="missing", admin_username="admin@example.com")
        assert result.status_code == 404


class TestAdditionalTrashBehaviour:
    """Extra tests to improve coverage for edge cases and cleanup logic."""

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash.fetch_all_experiments")
    async def test_list_deleted_experiments_handles_none_tags(self, mock_fetch_all_experiments):
        mock_deleted_experiment = MagicMock()
        mock_deleted_experiment.experiment_id = "321"
        mock_deleted_experiment.name = "Deleted No Tags"
        mock_deleted_experiment.lifecycle_stage = "deleted"
        mock_deleted_experiment.artifact_location = "/tmp/artifacts/321"
        mock_deleted_experiment.tags = None
        mock_deleted_experiment.creation_time = 10
        mock_deleted_experiment.last_update_time = 20

        mock_fetch_all_experiments.return_value = [mock_deleted_experiment]

        result = await list_deleted_experiments(admin_username="admin@example.com")
        assert result.status_code == 200
        import json

        payload = json.loads(result.body)
        assert payload["deleted_experiments"][0]["tags"] == {}

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash.fetch_all_experiments")
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_list_deleted_runs_fallback_and_empty(self, mock_get_store, mock_fetch_all_experiments):
        # Backend lacks _get_deleted_runs and search_runs raises -> fallback yields empty runs
        backend_store = MagicMock()
        if hasattr(backend_store, "_get_deleted_runs"):
            delattr(backend_store, "_get_deleted_runs")
        backend_store.search_runs.side_effect = Exception("search failed")
        mock_get_store.return_value = backend_store

        # Make fetch_all_experiments return one experiment id
        exp = MagicMock()
        exp.experiment_id = "exp-1"
        mock_fetch_all_experiments.return_value = [exp]

        result = await list_deleted_runs(admin_username="admin@example.com", experiment_ids=None, older_than=None)
        assert result.status_code == 200
        import json

        payload = json.loads(result.body)
        assert payload["deleted_runs"] == []

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_list_deleted_runs_skips_unfetchable_run(self, mock_get_store):
        backend_store = MagicMock()
        backend_store._get_deleted_runs.return_value = ["r-1"]
        backend_store.get_run.side_effect = Exception("unfetchable")
        mock_get_store.return_value = backend_store

        result = await list_deleted_runs(admin_username="admin@example.com", experiment_ids=None, older_than=None)
        assert result.status_code == 200
        import json

        payload = json.loads(result.body)
        assert payload["deleted_runs"] == []

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_cleanup_backend_without_hard_delete_run(self, mock_get_store):
        backend_store = MagicMock()
        # Remove _hard_delete_run capability
        if hasattr(backend_store, "_hard_delete_run"):
            delattr(backend_store, "_hard_delete_run")
        mock_get_store.return_value = backend_store

        result = await permanently_delete_all_trashed_entities(admin_username="admin@example.com", older_than=None)
        assert result.status_code == 400
        import json

        payload = json.loads(result.body)
        assert "Backend store does not support permanent deletion of runs" in payload["error"]

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_cleanup_experiment_not_found_returns_404(self, mock_get_store):
        backend_store = MagicMock()
        backend_store._hard_delete_run = MagicMock()
        backend_store._hard_delete_experiment = MagicMock()
        backend_store.get_experiment.side_effect = Exception("missing")
        mock_get_store.return_value = backend_store

        result = await permanently_delete_all_trashed_entities(
            admin_username="admin@example.com",
            run_ids=None,
            experiment_ids="nope",
            older_than=None,
        )
        assert result.status_code == 404
        import json

        payload = json.loads(result.body)
        assert "Experiment nope not found" in payload["error"]

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_cleanup_experiment_active_returns_400(self, mock_get_store):
        backend_store = MagicMock()
        backend_store._hard_delete_run = MagicMock()
        backend_store._hard_delete_experiment = MagicMock()

        active = MagicMock()
        active.lifecycle_stage = "active"
        active.experiment_id = "a1"
        backend_store.get_experiment.return_value = active
        mock_get_store.return_value = backend_store

        result = await permanently_delete_all_trashed_entities(
            admin_username="admin@example.com",
            run_ids=None,
            experiment_ids="a1",
            older_than=None,
        )
        assert result.status_code == 400
        import json

        payload = json.loads(result.body)
        assert "are not in deleted lifecycle stage" in payload["error"]

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash.get_artifact_repository")
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_cleanup_delete_runs_and_experiments_happy_path(self, mock_get_store, mock_get_artifact_repo):
        backend_store = MagicMock()
        backend_store._hard_delete_run = MagicMock()
        backend_store._hard_delete_experiment = MagicMock()

        # Setup run to be deleted
        run = MagicMock()
        run.info.run_id = "run-1"
        run.info.lifecycle_stage = "deleted"
        run.info.artifact_uri = "invalid://"
        run.info.experiment_id = "exp-1"
        backend_store.get_run.return_value = run

        # _get_deleted_runs returns our run id
        backend_store._get_deleted_runs.return_value = ["run-1"]

        # Experiment to delete
        exp = MagicMock()
        exp.experiment_id = "exp-1"
        exp.lifecycle_stage = "deleted"
        exp.last_update_time = 0

        backend_store.get_experiment.return_value = exp

        # Make artifact repo deletion raise InvalidUrlException to exercise that branch
        from mlflow.exceptions import InvalidUrlException

        mock_repo = MagicMock()
        mock_repo.delete_artifacts.side_effect = InvalidUrlException("bad url")
        mock_get_artifact_repo.return_value = mock_repo

        mock_get_store.return_value = backend_store

        result = await permanently_delete_all_trashed_entities(
            admin_username="admin@example.com",
            run_ids="run-1",
            experiment_ids="exp-1",
            older_than=None,
        )
        assert result.status_code == 200
        import json

        payload = json.loads(result.body)
        assert payload["deleted_runs"] == ["run-1"]
        assert payload["deleted_experiments"] == ["exp-1"]

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_cleanup_run_not_deleted_and_hard_delete_experiment_failure(self, mock_get_store):
        backend_store = MagicMock()
        backend_store._hard_delete_run = MagicMock()

        # Run exists but is active
        run = MagicMock()
        run.info.run_id = "run-2"
        run.info.lifecycle_stage = "active"
        backend_store.get_run.return_value = run

        # Experiment deletion will raise
        exp = MagicMock()
        exp.experiment_id = "e2"
        exp.lifecycle_stage = "deleted"
        backend_store.get_experiment.return_value = exp
        backend_store._hard_delete_experiment.side_effect = Exception("boom")

        mock_get_store.return_value = backend_store

        result = await permanently_delete_all_trashed_entities(
            admin_username="admin@example.com",
            run_ids="run-2",
            experiment_ids="e2",
            older_than=None,
        )
        assert result.status_code == 200
        import json

        payload = json.loads(result.body)
        # run should not be deleted and should appear in failed_runs
        assert any(f["run_id"] == "run-2" for f in payload.get("failed_runs", []))
        # experiment deletion should have failed
        assert any(f["experiment_id"] == "e2" for f in payload.get("failed_experiments", []))

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_list_deleted_runs_filters_by_experiment(self, mock_get_store):
        backend_store = MagicMock()
        backend_store._get_deleted_runs.return_value = ["r1", "r2"]

        run1 = MagicMock()
        run1.info.run_id = "r1"
        run1.info.experiment_id = "exp-1"
        run1.info.lifecycle_stage = "deleted"
        run1.info.run_name = "n1"
        run1.info.status = "FINISHED"
        run1.info.start_time = 1
        run1.info.end_time = 2

        run2 = MagicMock()
        run2.info.run_id = "r2"
        run2.info.experiment_id = "exp-2"
        run2.info.lifecycle_stage = "deleted"
        run2.info.run_name = "n2"
        run2.info.status = "FINISHED"
        run2.info.start_time = 3
        run2.info.end_time = 4

        backend_store.get_run.side_effect = [run1, run2]
        mock_get_store.return_value = backend_store

        result = await list_deleted_runs(admin_username="admin@example.com", experiment_ids="exp-1", older_than=None)
        assert result.status_code == 200
        import json

        payload = json.loads(result.body)
        assert payload["deleted_runs"] == [
            {
                "run_id": "r1",
                "experiment_id": "exp-1",
                "run_name": "n1",
                "status": "FINISHED",
                "start_time": 1,
                "end_time": 2,
                "lifecycle_stage": "deleted",
            }
        ]

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_list_deleted_runs_backend_raises_returns_500(self, mock_get_store):
        backend_store = MagicMock()
        backend_store._get_deleted_runs.side_effect = Exception("boom")
        mock_get_store.return_value = backend_store

        from fastapi import HTTPException

        with pytest.raises(HTTPException) as excinfo:
            await list_deleted_runs(admin_username="admin@example.com", experiment_ids=None, older_than=None)

        assert excinfo.value.status_code == 500
        assert excinfo.value.detail == "Failed to retrieve deleted runs"

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash.get_artifact_repository")
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_cleanup_fetch_experiments_and_runs(self, mock_get_store, mock_get_artifact_repo):
        backend_store = MagicMock()
        backend_store._hard_delete_run = MagicMock()
        backend_store._hard_delete_experiment = MagicMock()

        # Implement a simple Page class used by backend search methods
        class Page(list):
            def __init__(self, items, token=None):
                super().__init__(items)
                self.token = token

            def __add__(self, other):
                return Page(list(self) + list(other), token=None)

        # Experiment returned by search_experiments
        exp = MagicMock()
        exp.experiment_id = "e1"
        exp.lifecycle_stage = "deleted"
        exp.last_update_time = 0

        # Run returned by search_runs
        run = MagicMock()
        run.info.run_id = "r1"
        run.info.lifecycle_stage = "deleted"
        run.info.artifact_uri = "invalid://"
        run.info.experiment_id = "e1"
        run.info.run_name = "rname"
        run.info.status = "FINISHED"
        run.info.start_time = 1
        run.info.end_time = 2

        backend_store.search_experiments.return_value = Page([exp], token=None)
        backend_store.search_runs.return_value = Page([run], token=None)
        backend_store._get_deleted_runs.return_value = ["r1"]
        backend_store.get_run.return_value = run

        # artifact repo raising InvalidUrl triggers warning branch but doesn't fail
        from mlflow.exceptions import InvalidUrlException

        mock_repo = MagicMock()
        mock_repo.delete_artifacts.side_effect = InvalidUrlException("bad url")
        mock_get_artifact_repo.return_value = mock_repo

        mock_get_store.return_value = backend_store

        result = await permanently_delete_all_trashed_entities(
            admin_username="admin@example.com",
            older_than=None,
            run_ids=None,
            experiment_ids=None,
        )
        assert result.status_code == 200
        import json

        payload = json.loads(result.body)
        assert payload["deleted_runs"] == ["r1"]
        assert payload["deleted_experiments"] == ["e1"]

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_list_deleted_runs_paged_search_runs(self, mock_get_store):
        backend_store = MagicMock()

        # Implement a Page that supports token and addition
        class Page(list):
            def __init__(self, items, token=None):
                super().__init__(items)
                self.token = token

            def __add__(self, other):
                return Page(list(self) + list(other), token=None)

        # two runs across pages
        run1 = MagicMock()
        run1.info.run_id = "r1"
        run1.info.experiment_id = "exp-1"
        run1.info.lifecycle_stage = "deleted"
        run1.info.run_name = "n1"
        run1.info.status = "FINISHED"
        run1.info.start_time = 1
        run1.info.end_time = 2

        run2 = MagicMock()
        run2.info.run_id = "r2"
        run2.info.experiment_id = "exp-2"
        run2.info.lifecycle_stage = "deleted"
        run2.info.run_name = "n2"
        run2.info.status = "FINISHED"
        run2.info.start_time = 3
        run2.info.end_time = 4

        def search_runs(experiment_ids, filter_string, run_view_type, page_token=None):
            if page_token is None:
                return Page([run1], token="t")
            else:
                return Page([run2], token=None)

        backend_store.search_runs.side_effect = search_runs
        # Ensure backend_store has no _get_deleted_runs attribute so fallback path is used
        if hasattr(backend_store, "_get_deleted_runs"):
            delattr(backend_store, "_get_deleted_runs")
        # fetch_all_experiments used when experiment_ids not provided
        with patch("mlflow_oidc_auth.routers.trash.fetch_all_experiments") as mock_fetch_all_experiments:
            exp = MagicMock()
            exp.experiment_id = "exp-1"
            mock_fetch_all_experiments.return_value = [exp]

            backend_store.get_run.side_effect = [run1, run2]
            mock_get_store.return_value = backend_store

            result = await list_deleted_runs(admin_username="admin@example.com", experiment_ids=None, older_than=None)
            assert result.status_code == 200
            import json

            payload = json.loads(result.body)
            assert len(payload["deleted_runs"]) == 2

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_cleanup_run_not_old_returns_failed_run(self, mock_get_store):
        backend_store = MagicMock()
        backend_store._hard_delete_run = MagicMock()
        # no deleted runs older than
        backend_store._get_deleted_runs.return_value = []

        # run exists and is deleted
        run = MagicMock()
        run.info.run_id = "r1"
        run.info.lifecycle_stage = "deleted"
        run.info.artifact_uri = "invalid://"
        run.info.experiment_id = "e1"
        run.info.run_name = "r"
        run.info.status = "FINISHED"
        run.info.start_time = 1
        run.info.end_time = 2

        backend_store.get_run.return_value = run
        mock_get_store.return_value = backend_store

        result = await permanently_delete_all_trashed_entities(
            admin_username="admin@example.com",
            older_than="1d",
            run_ids="r1",
            experiment_ids=None,
        )
        assert result.status_code == 200
        import json

        payload = json.loads(result.body)
        assert any(f["run_id"] == "r1" for f in payload.get("failed_runs", []))

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_cleanup_experiment_age_check_non_old(self, mock_get_store):
        backend_store = MagicMock()
        backend_store._hard_delete_run = MagicMock()
        backend_store._hard_delete_experiment = MagicMock()

        # experiment provided and last_update_time recent
        active_exp = MagicMock()
        active_exp.experiment_id = "eX"
        active_exp.lifecycle_stage = "deleted"
        # set last_update_time to current to be "non old"
        from mlflow.utils.time import get_current_time_millis

        active_exp.last_update_time = get_current_time_millis()

        backend_store.get_experiment.return_value = active_exp
        mock_get_store.return_value = backend_store

        result = await permanently_delete_all_trashed_entities(
            admin_username="admin@example.com",
            older_than="1d",
            run_ids=None,
            experiment_ids="eX",
        )
        assert result.status_code == 400
        import json

        payload = json.loads(result.body)
        assert "not older than" in payload["error"]

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash.get_artifact_repository")
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_cleanup_fetch_experiments_pages_and_runs(self, mock_get_store, mock_get_artifact_repo):
        backend_store = MagicMock()
        backend_store._hard_delete_run = MagicMock()
        backend_store._hard_delete_experiment = MagicMock()

        class Page(list):
            def __init__(self, items, token=None):
                super().__init__(items)
                self.token = token

            def __add__(self, other):
                return Page(list(self) + list(other), token=None)

        exp1 = MagicMock()
        exp1.experiment_id = "e1"
        exp2 = MagicMock()
        exp2.experiment_id = "e2"

        run1 = MagicMock()
        run1.info.run_id = "r1"
        run1.info.lifecycle_stage = "deleted"
        run1.info.artifact_uri = "invalid://"
        run1.info.experiment_id = "e1"
        run1.info.run_name = "r"
        run1.info.status = "FINISHED"
        run1.info.start_time = 1
        run1.info.end_time = 2
        run2 = MagicMock()
        run2.info.run_id = "r2"
        run2.info.lifecycle_stage = "deleted"
        run2.info.artifact_uri = "invalid://"
        run2.info.experiment_id = "e2"
        run2.info.run_name = "r"
        run2.info.status = "FINISHED"
        run2.info.start_time = 1
        run2.info.end_time = 2

        # make search_experiments return two pages
        def search_experiments(view_type, filter_string, page_token=None):
            if page_token is None:
                return Page([exp1], token="t")
            return Page([exp2], token=None)

        # make search_runs return page per experiments group
        def search_runs(experiment_ids, filter_string, run_view_type, page_token=None):
            if experiment_ids == ["e1"]:
                return Page([run1], token=None)
            if experiment_ids == ["e2"]:
                return Page([run2], token=None)
            return Page([], token=None)

        backend_store.search_experiments.side_effect = search_experiments
        backend_store.search_runs.side_effect = search_runs
        backend_store._get_deleted_runs.return_value = []
        backend_store.get_run.side_effect = [run1, run2]

        from mlflow.exceptions import InvalidUrlException

        mock_repo = MagicMock()
        mock_repo.delete_artifacts.side_effect = InvalidUrlException("bad url")
        mock_get_artifact_repo.return_value = mock_repo

        mock_get_store.return_value = backend_store

        result = await permanently_delete_all_trashed_entities(
            admin_username="admin@example.com",
            older_than=None,
            run_ids=None,
            experiment_ids=None,
        )
        assert result.status_code == 200
        import json

        payload = json.loads(result.body)
        assert set(payload["deleted_experiments"]) == {"e1", "e2"}

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_deleted_runs_fetch_failure_is_handled(self, mock_get_store):
        backend_store = MagicMock()
        backend_store._hard_delete_run = MagicMock()
        # _get_deleted_runs raises
        backend_store._get_deleted_runs.side_effect = Exception("boom-fetch")
        mock_get_store.return_value = backend_store

        result = await permanently_delete_all_trashed_entities(
            admin_username="admin@example.com",
            older_than=None,
            run_ids=None,
            experiment_ids=None,
        )
        assert result.status_code == 200
        import json

        payload = json.loads(result.body)
        assert payload["deleted_runs"] == []
        assert payload["deleted_experiments"] == []

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_cleanup_older_than_invalid_returns_400(self, mock_get_store):
        backend_store = MagicMock()
        backend_store._hard_delete_run = MagicMock()
        mock_get_store.return_value = backend_store

        result = await permanently_delete_all_trashed_entities(
            admin_username="admin@example.com",
            older_than="bad",
            run_ids=None,
            experiment_ids=None,
        )
        assert result.status_code == 400

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_list_deleted_runs_json_serialization_error_raises_http_exception(self, mock_get_store):
        backend_store = MagicMock()
        backend_store._get_deleted_runs.return_value = ["r1"]

        run = MagicMock()
        run.info.run_id = "r1"
        run.info.experiment_id = "exp-1"
        run.info.lifecycle_stage = "deleted"
        # Non-serializable field
        run.info.run_name = MagicMock()
        run.info.status = MagicMock()
        run.info.start_time = MagicMock()
        run.info.end_time = MagicMock()

        backend_store.get_run.return_value = run
        mock_get_store.return_value = backend_store

        from fastapi import HTTPException

        with pytest.raises(HTTPException) as excinfo:
            await list_deleted_runs(admin_username="admin@example.com", experiment_ids=None, older_than=None)

        assert excinfo.value.status_code == 500
        assert excinfo.value.detail == "Failed to retrieve deleted runs"

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash.get_artifact_repository")
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_cleanup_artifact_delete_exception_is_handled(self, mock_get_store, mock_get_artifact_repo):
        backend_store = MagicMock()
        backend_store._hard_delete_run = MagicMock()
        backend_store._hard_delete_experiment = MagicMock()
        backend_store._get_deleted_runs.return_value = ["r1"]

        run = MagicMock()
        run.info.run_id = "r1"
        run.info.lifecycle_stage = "deleted"
        run.info.artifact_uri = "some://"
        run.info.experiment_id = "e1"

        backend_store.get_run.return_value = run
        mock_get_store.return_value = backend_store

        mock_repo = MagicMock()
        mock_repo.delete_artifacts.side_effect = Exception("boom-artifact")
        mock_get_artifact_repo.return_value = mock_repo

        result = await permanently_delete_all_trashed_entities(
            admin_username="admin@example.com",
            run_ids="r1",
            experiment_ids=None,
            older_than=None,
        )
        assert result.status_code == 200
        import json

        payload = json.loads(result.body)
        assert payload["deleted_runs"] == ["r1"]

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash.get_artifact_repository")
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_cleanup_hard_delete_run_failure_records_failed_run(self, mock_get_store, mock_get_artifact_repo):
        backend_store = MagicMock()
        # hard delete run will fail
        backend_store._hard_delete_run.side_effect = Exception("boom-delete")
        backend_store._get_deleted_runs.return_value = ["r1"]

        run = MagicMock()
        run.info.run_id = "r1"
        run.info.lifecycle_stage = "deleted"
        run.info.artifact_uri = "some://"
        run.info.experiment_id = "e1"

        backend_store.get_run.return_value = run
        mock_get_store.return_value = backend_store

        mock_repo = MagicMock()
        mock_repo.delete_artifacts.return_value = None
        mock_get_artifact_repo.return_value = mock_repo

        result = await permanently_delete_all_trashed_entities(
            admin_username="admin@example.com",
            run_ids="r1",
            experiment_ids=None,
            older_than=None,
        )
        assert result.status_code == 200
        import json

        payload = json.loads(result.body)
        assert any(f["run_id"] == "r1" and "boom-delete" in f["error"] for f in payload.get("failed_runs", []))

    def test_parse_time_delta_more_cases(self):
        from mlflow.exceptions import MlflowException

        from mlflow_oidc_auth.routers.trash import _parse_time_delta

        assert _parse_time_delta("1.5h") == int(1.5 * 3600 * 1000)
        assert _parse_time_delta("2d8h5m20s") == int((2 * 24 * 3600 + 8 * 3600 + 5 * 60 + 20) * 1000)

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_cleanup_skips_experiments_when_not_supported(self, mock_get_store):
        backend_store = MagicMock()
        backend_store._hard_delete_run = MagicMock()
        # remove _hard_delete_experiment if present
        if hasattr(backend_store, "_hard_delete_experiment"):
            delattr(backend_store, "_hard_delete_experiment")
        # No runs or experiments to delete
        backend_store._get_deleted_runs.return_value = []
        mock_get_store.return_value = backend_store

        import warnings

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = await permanently_delete_all_trashed_entities(
                admin_username="admin@example.com",
                older_than=None,
                run_ids=None,
                experiment_ids=None,
            )
            assert result.status_code == 200
            import json

            payload = json.loads(result.body)
            assert payload["deleted_runs"] == []
            assert payload["deleted_experiments"] == []
            # Ensure we warned about experiments not supported
            assert any("does not allow hard-deleting experiments" in str(x.message) for x in w)

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_cleanup_top_level_exception_raises_http_exception(self, mock_get_store):
        mock_get_store.side_effect = Exception("boom")
        from fastapi import HTTPException

        with pytest.raises(HTTPException) as excinfo:
            await permanently_delete_all_trashed_entities(
                admin_username="admin@example.com",
                older_than=None,
                run_ids=None,
                experiment_ids=None,
            )

        assert excinfo.value.status_code == 500
        assert excinfo.value.detail == "Cleanup operation failed"

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_restore_experiment_fails_raises_500(self, mock_get_store):
        backend_store = MagicMock()
        deleted = MagicMock()
        deleted.lifecycle_stage = "deleted"
        deleted.experiment_id = "999"
        deleted.name = "exp"
        deleted.last_update_time = 1

        backend_store.get_experiment.return_value = deleted
        backend_store.restore_experiment.side_effect = Exception("boom")
        mock_get_store.return_value = backend_store

        from fastapi import HTTPException

        with pytest.raises(HTTPException) as excinfo:
            await restore_experiment(experiment_id="999", admin_username="admin@example.com")

        assert excinfo.value.status_code == 500
        assert excinfo.value.detail == "Failed to restore experiment"

    @pytest.mark.asyncio
    @patch("mlflow_oidc_auth.routers.trash._get_store")
    async def test_restore_run_fails_raises_500(self, mock_get_store):
        backend_store = MagicMock()
        deleted = MagicMock()
        deleted.info.lifecycle_stage = "deleted"
        deleted.info.run_id = "run-9"
        deleted.info.experiment_id = "exp-9"
        deleted.info.run_name = "r"
        deleted.info.status = "FINISHED"

        backend_store.get_run.return_value = deleted
        backend_store.restore_run.side_effect = Exception("boom")
        mock_get_store.return_value = backend_store

        from fastapi import HTTPException

        with pytest.raises(HTTPException) as excinfo:
            await restore_run(run_id="run-9", admin_username="admin@example.com")

        assert excinfo.value.status_code == 500
        assert excinfo.value.detail == "Failed to restore run"

    def test_parse_time_delta_valid_and_invalid(self):
        from mlflow.exceptions import MlflowException

        from mlflow_oidc_auth.routers.trash import _parse_time_delta, _split_csv

        # valid
        assert _parse_time_delta("1s") == 1000
        # invalid
        with pytest.raises(MlflowException):
            _parse_time_delta("bad")

        # _split_csv
        assert _split_csv(None) == []
        assert _split_csv("a, b, ,c") == ["a", "b", "c"]



"""Tests for the batch permissions module.

These tests verify the batch permission resolution functions work correctly
by mocking the store methods and verifying the permission resolution logic.
"""

from unittest.mock import MagicMock, patch

import pytest

from mlflow_oidc_auth.utils.batch_permissions import (
    UserPermissionContext,
    build_user_permission_context,
    resolve_experiment_permission_from_context,
    resolve_model_permission_from_context,
    resolve_prompt_permission_from_context,
    batch_resolve_experiment_permissions,
    batch_resolve_model_permissions,
    batch_resolve_prompt_permissions,
    filter_manageable_experiments,
    filter_manageable_models,
    filter_manageable_prompts,
    _find_regex_permission,
    _resolve_permission_from_context,
)


class TestFindRegexPermission:
    """Tests for the _find_regex_permission helper function."""

    def test_finds_matching_regex(self):
        """Should return permission when regex matches."""
        regex_perm = MagicMock()
        regex_perm.regex = "^model-.*"
        regex_perm.permission = "READ"

        result = _find_regex_permission([regex_perm], "model-test")
        assert result == "READ"

    def test_returns_none_when_no_match(self):
        """Should return None when no regex matches."""
        regex_perm = MagicMock()
        regex_perm.regex = "^other-.*"
        regex_perm.permission = "READ"

        result = _find_regex_permission([regex_perm], "model-test")
        assert result is None

    def test_returns_first_matching_regex(self):
        """Should return the first matching regex permission."""
        regex_perm1 = MagicMock()
        regex_perm1.regex = ".*"
        regex_perm1.permission = "READ"

        regex_perm2 = MagicMock()
        regex_perm2.regex = "^model-.*"
        regex_perm2.permission = "MANAGE"

        result = _find_regex_permission([regex_perm1, regex_perm2], "model-test")
        assert result == "READ"

    def test_empty_list_returns_none(self):
        """Should return None for empty regex list."""
        result = _find_regex_permission([], "model-test")
        assert result is None

    def test_partial_match_with_anchored_regex(self):
        """Should only match if regex pattern matches."""
        regex_perm = MagicMock()
        regex_perm.regex = "^exact$"
        regex_perm.permission = "MANAGE"

        assert _find_regex_permission([regex_perm], "exact") == "MANAGE"
        assert _find_regex_permission([regex_perm], "exact-extra") is None
        assert _find_regex_permission([regex_perm], "not-exact") is None


class TestResolvePermissionFromContext:
    """Tests for the _resolve_permission_from_context helper function."""

    def test_returns_user_permission_first(self):
        """Should return user permission when it's first in source order."""
        with patch("mlflow_oidc_auth.utils.batch_permissions.config") as mock_config:
            mock_config.PERMISSION_SOURCE_ORDER = [
                "user",
                "group",
                "regex",
                "group-regex",
            ]
            mock_config.DEFAULT_MLFLOW_PERMISSION = "NO_PERMISSIONS"

            result = _resolve_permission_from_context(
                mock_config.PERMISSION_SOURCE_ORDER,
                user_direct="MANAGE",
                group_direct="READ",
                user_regex_match="EDIT",
                group_regex_match="READ",
            )

            assert result.permission.name == "MANAGE"
            assert result.kind == "user"

    def test_returns_group_permission_when_no_user(self):
        """Should return group permission when user permission is None."""
        with patch("mlflow_oidc_auth.utils.batch_permissions.config") as mock_config:
            mock_config.PERMISSION_SOURCE_ORDER = [
                "user",
                "group",
                "regex",
                "group-regex",
            ]
            mock_config.DEFAULT_MLFLOW_PERMISSION = "NO_PERMISSIONS"

            result = _resolve_permission_from_context(
                mock_config.PERMISSION_SOURCE_ORDER,
                user_direct=None,
                group_direct="READ",
                user_regex_match="EDIT",
                group_regex_match=None,
            )

            assert result.permission.name == "READ"
            assert result.kind == "group"

    def test_returns_regex_permission_when_no_direct(self):
        """Should return regex permission when no direct permissions."""
        with patch("mlflow_oidc_auth.utils.batch_permissions.config") as mock_config:
            mock_config.PERMISSION_SOURCE_ORDER = [
                "user",
                "group",
                "regex",
                "group-regex",
            ]
            mock_config.DEFAULT_MLFLOW_PERMISSION = "NO_PERMISSIONS"

            result = _resolve_permission_from_context(
                mock_config.PERMISSION_SOURCE_ORDER,
                user_direct=None,
                group_direct=None,
                user_regex_match="EDIT",
                group_regex_match="READ",
            )

            assert result.permission.name == "EDIT"
            assert result.kind == "regex"

    def test_returns_group_regex_permission_when_no_user_regex(self):
        """Should return group-regex permission when no user regex."""
        with patch("mlflow_oidc_auth.utils.batch_permissions.config") as mock_config:
            mock_config.PERMISSION_SOURCE_ORDER = [
                "user",
                "group",
                "regex",
                "group-regex",
            ]
            mock_config.DEFAULT_MLFLOW_PERMISSION = "NO_PERMISSIONS"

            result = _resolve_permission_from_context(
                mock_config.PERMISSION_SOURCE_ORDER,
                user_direct=None,
                group_direct=None,
                user_regex_match=None,
                group_regex_match="MANAGE",
            )

            assert result.permission.name == "MANAGE"
            assert result.kind == "group-regex"

    def test_returns_default_permission_when_no_match(self):
        """Should return default permission when no sources match."""
        with patch("mlflow_oidc_auth.utils.batch_permissions.config") as mock_config:
            mock_config.PERMISSION_SOURCE_ORDER = [
                "user",
                "group",
                "regex",
                "group-regex",
            ]
            mock_config.DEFAULT_MLFLOW_PERMISSION = "READ"

            result = _resolve_permission_from_context(
                mock_config.PERMISSION_SOURCE_ORDER,
                user_direct=None,
                group_direct=None,
                user_regex_match=None,
                group_regex_match=None,
            )

            assert result.permission.name == "READ"
            assert result.kind == "fallback"

    def test_respects_custom_source_order(self):
        """Should respect custom source order configuration."""
        with patch("mlflow_oidc_auth.utils.batch_permissions.config") as mock_config:
            # Group first
            mock_config.PERMISSION_SOURCE_ORDER = [
                "group",
                "user",
                "regex",
                "group-regex",
            ]
            mock_config.DEFAULT_MLFLOW_PERMISSION = "NO_PERMISSIONS"

            result = _resolve_permission_from_context(
                mock_config.PERMISSION_SOURCE_ORDER,
                user_direct="MANAGE",
                group_direct="READ",
                user_regex_match=None,
                group_regex_match=None,
            )

            assert result.permission.name == "READ"
            assert result.kind == "group"


class TestUserPermissionContext:
    """Tests for the UserPermissionContext dataclass."""

    def test_dataclass_creation(self):
        """Should create dataclass with all fields."""
        ctx = UserPermissionContext(
            username="testuser",
            group_ids=[1, 2, 3],
            user_experiment_permissions={"exp-1": "READ"},
            group_experiment_permissions={"exp-2": "MANAGE"},
            experiment_regex_permissions=[],
            group_experiment_regex_permissions=[],
            user_model_permissions={"model-a": "EDIT"},
            group_model_permissions={},
            model_regex_permissions=[],
            group_model_regex_permissions=[],
            prompt_regex_permissions=[],
            group_prompt_regex_permissions=[],
        )

        assert ctx.username == "testuser"
        assert ctx.group_ids == [1, 2, 3]
        assert ctx.user_experiment_permissions == {"exp-1": "READ"}
        assert ctx.group_experiment_permissions == {"exp-2": "MANAGE"}
        assert ctx.user_model_permissions == {"model-a": "EDIT"}


class TestBuildUserPermissionContext:
    """Tests for the build_user_permission_context function."""

    @patch("mlflow_oidc_auth.utils.batch_permissions.store")
    def test_builds_context_with_group_ids(self, mock_store):
        """Should build context fetching all permission data when user has groups."""
        # Setup mocks
        mock_store.get_groups_ids_for_user.return_value = [1, 2]

        exp_perm = MagicMock()
        exp_perm.experiment_id = "exp-1"
        exp_perm.permission = "READ"
        mock_store.list_experiment_permissions.return_value = [exp_perm]

        group_exp_perm = MagicMock()
        group_exp_perm.experiment_id = "exp-2"
        group_exp_perm.permission = "MANAGE"
        mock_store.list_user_groups_experiment_permissions.return_value = [group_exp_perm]

        mock_store.list_experiment_regex_permissions.return_value = []
        mock_store.list_group_experiment_regex_permissions_for_groups_ids.return_value = []

        model_perm = MagicMock()
        model_perm.name = "model-a"
        model_perm.permission = "EDIT"
        mock_store.list_registered_model_permissions.return_value = [model_perm]

        mock_store.list_user_groups_registered_model_permissions.return_value = []
        mock_store.list_registered_model_regex_permissions.return_value = []
        mock_store.list_group_registered_model_regex_permissions_for_groups_ids.return_value = []
        mock_store.list_prompt_regex_permissions.return_value = []
        mock_store.list_group_prompt_regex_permissions_for_groups_ids.return_value = []

        ctx = build_user_permission_context("testuser")

        assert ctx.username == "testuser"
        assert ctx.group_ids == [1, 2]
        assert ctx.user_experiment_permissions == {"exp-1": "READ"}
        assert ctx.group_experiment_permissions == {"exp-2": "MANAGE"}
        assert ctx.user_model_permissions == {"model-a": "EDIT"}

        # Verify store calls
        mock_store.get_groups_ids_for_user.assert_called_once_with("testuser")
        mock_store.list_experiment_permissions.assert_called_once_with("testuser")
        mock_store.list_group_experiment_regex_permissions_for_groups_ids.assert_called_once_with([1, 2])

    @patch("mlflow_oidc_auth.utils.batch_permissions.store")
    def test_builds_context_without_groups(self, mock_store):
        """Should skip group regex queries when user has no groups."""
        mock_store.get_groups_ids_for_user.return_value = []
        mock_store.list_experiment_permissions.return_value = []
        mock_store.list_user_groups_experiment_permissions.return_value = []
        mock_store.list_experiment_regex_permissions.return_value = []
        mock_store.list_registered_model_permissions.return_value = []
        mock_store.list_user_groups_registered_model_permissions.return_value = []
        mock_store.list_registered_model_regex_permissions.return_value = []
        mock_store.list_prompt_regex_permissions.return_value = []

        ctx = build_user_permission_context("testuser")

        assert ctx.group_ids == []
        assert ctx.group_experiment_regex_permissions == []
        assert ctx.group_model_regex_permissions == []
        assert ctx.group_prompt_regex_permissions == []

        # Should NOT call group regex methods when no groups
        mock_store.list_group_experiment_regex_permissions_for_groups_ids.assert_not_called()
        mock_store.list_group_registered_model_regex_permissions_for_groups_ids.assert_not_called()
        mock_store.list_group_prompt_regex_permissions_for_groups_ids.assert_not_called()


class TestResolveExperimentPermissionFromContext:
    """Tests for resolving experiment permissions from context."""

    @pytest.fixture
    def context_with_experiment_permissions(self):
        """Create context with experiment permissions."""
        exp_regex = MagicMock()
        exp_regex.regex = "^test-.*"
        exp_regex.permission = "EDIT"

        group_regex = MagicMock()
        group_regex.regex = "^prod-.*"
        group_regex.permission = "READ"

        return UserPermissionContext(
            username="testuser",
            group_ids=[1],
            user_experiment_permissions={"exp-1": "MANAGE"},
            group_experiment_permissions={"exp-2": "READ"},
            experiment_regex_permissions=[exp_regex],
            group_experiment_regex_permissions=[group_regex],
            user_model_permissions={},
            group_model_permissions={},
            model_regex_permissions=[],
            group_model_regex_permissions=[],
            prompt_regex_permissions=[],
            group_prompt_regex_permissions=[],
        )

    def test_resolves_user_experiment_permission(self, context_with_experiment_permissions):
        """Should resolve user direct experiment permission."""
        with patch("mlflow_oidc_auth.utils.batch_permissions.config") as mock_config:
            mock_config.PERMISSION_SOURCE_ORDER = [
                "user",
                "group",
                "regex",
                "group-regex",
            ]
            mock_config.DEFAULT_MLFLOW_PERMISSION = "NO_PERMISSIONS"

            result = resolve_experiment_permission_from_context(context_with_experiment_permissions, "exp-1", "experiment-name")

            assert result.permission.name == "MANAGE"
            assert result.kind == "user"

    def test_resolves_group_experiment_permission(self, context_with_experiment_permissions):
        """Should resolve group experiment permission when no user permission."""
        with patch("mlflow_oidc_auth.utils.batch_permissions.config") as mock_config:
            mock_config.PERMISSION_SOURCE_ORDER = [
                "user",
                "group",
                "regex",
                "group-regex",
            ]
            mock_config.DEFAULT_MLFLOW_PERMISSION = "NO_PERMISSIONS"

            result = resolve_experiment_permission_from_context(context_with_experiment_permissions, "exp-2", "experiment-name")

            assert result.permission.name == "READ"
            assert result.kind == "group"

    def test_resolves_regex_experiment_permission(self, context_with_experiment_permissions):
        """Should resolve regex experiment permission by experiment name."""
        with patch("mlflow_oidc_auth.utils.batch_permissions.config") as mock_config:
            mock_config.PERMISSION_SOURCE_ORDER = [
                "user",
                "group",
                "regex",
                "group-regex",
            ]
            mock_config.DEFAULT_MLFLOW_PERMISSION = "NO_PERMISSIONS"

            result = resolve_experiment_permission_from_context(context_with_experiment_permissions, "exp-unknown", "test-experiment")

            assert result.permission.name == "EDIT"
            assert result.kind == "regex"

    def test_resolves_group_regex_experiment_permission(self, context_with_experiment_permissions):
        """Should resolve group regex experiment permission."""
        with patch("mlflow_oidc_auth.utils.batch_permissions.config") as mock_config:
            mock_config.PERMISSION_SOURCE_ORDER = [
                "user",
                "group",
                "regex",
                "group-regex",
            ]
            mock_config.DEFAULT_MLFLOW_PERMISSION = "NO_PERMISSIONS"

            result = resolve_experiment_permission_from_context(context_with_experiment_permissions, "exp-unknown", "prod-experiment")

            assert result.permission.name == "READ"
            assert result.kind == "group-regex"

    @patch("mlflow_oidc_auth.utils.batch_permissions._get_tracking_store")
    def test_fetches_experiment_name_when_not_provided(self, mock_get_store, context_with_experiment_permissions):
        """Should fetch experiment name from tracking store when not provided."""
        mock_experiment = MagicMock()
        mock_experiment.name = "test-fetched-experiment"
        mock_get_store.return_value.get_experiment.return_value = mock_experiment

        with patch("mlflow_oidc_auth.utils.batch_permissions.config") as mock_config:
            mock_config.PERMISSION_SOURCE_ORDER = [
                "user",
                "group",
                "regex",
                "group-regex",
            ]
            mock_config.DEFAULT_MLFLOW_PERMISSION = "NO_PERMISSIONS"

            result = resolve_experiment_permission_from_context(context_with_experiment_permissions, "exp-unknown", None)

            mock_get_store.return_value.get_experiment.assert_called_once_with("exp-unknown")
            # Should match the regex "^test-.*"
            assert result.permission.name == "EDIT"


class TestResolveModelPermissionFromContext:
    """Tests for resolving model permissions from context."""

    @pytest.fixture
    def context_with_model_permissions(self):
        """Create context with model permissions."""
        model_regex = MagicMock()
        model_regex.regex = "^ml-.*"
        model_regex.permission = "EDIT"

        group_model_regex = MagicMock()
        group_model_regex.regex = "^shared-.*"
        group_model_regex.permission = "READ"

        return UserPermissionContext(
            username="testuser",
            group_ids=[1],
            user_experiment_permissions={},
            group_experiment_permissions={},
            experiment_regex_permissions=[],
            group_experiment_regex_permissions=[],
            user_model_permissions={"my-model": "MANAGE"},
            group_model_permissions={"team-model": "EDIT"},
            model_regex_permissions=[model_regex],
            group_model_regex_permissions=[group_model_regex],
            prompt_regex_permissions=[],
            group_prompt_regex_permissions=[],
        )

    def test_resolves_user_model_permission(self, context_with_model_permissions):
        """Should resolve user direct model permission."""
        with patch("mlflow_oidc_auth.utils.batch_permissions.config") as mock_config:
            mock_config.PERMISSION_SOURCE_ORDER = [
                "user",
                "group",
                "regex",
                "group-regex",
            ]
            mock_config.DEFAULT_MLFLOW_PERMISSION = "NO_PERMISSIONS"

            result = resolve_model_permission_from_context(context_with_model_permissions, "my-model")

            assert result.permission.name == "MANAGE"
            assert result.kind == "user"

    def test_resolves_group_model_permission(self, context_with_model_permissions):
        """Should resolve group model permission."""
        with patch("mlflow_oidc_auth.utils.batch_permissions.config") as mock_config:
            mock_config.PERMISSION_SOURCE_ORDER = [
                "user",
                "group",
                "regex",
                "group-regex",
            ]
            mock_config.DEFAULT_MLFLOW_PERMISSION = "NO_PERMISSIONS"

            result = resolve_model_permission_from_context(context_with_model_permissions, "team-model")

            assert result.permission.name == "EDIT"
            assert result.kind == "group"

    def test_resolves_regex_model_permission(self, context_with_model_permissions):
        """Should resolve regex model permission."""
        with patch("mlflow_oidc_auth.utils.batch_permissions.config") as mock_config:
            mock_config.PERMISSION_SOURCE_ORDER = [
                "user",
                "group",
                "regex",
                "group-regex",
            ]
            mock_config.DEFAULT_MLFLOW_PERMISSION = "NO_PERMISSIONS"

            result = resolve_model_permission_from_context(context_with_model_permissions, "ml-classifier")

            assert result.permission.name == "EDIT"
            assert result.kind == "regex"

    def test_resolves_group_regex_model_permission(self, context_with_model_permissions):
        """Should resolve group regex model permission."""
        with patch("mlflow_oidc_auth.utils.batch_permissions.config") as mock_config:
            mock_config.PERMISSION_SOURCE_ORDER = [
                "user",
                "group",
                "regex",
                "group-regex",
            ]
            mock_config.DEFAULT_MLFLOW_PERMISSION = "NO_PERMISSIONS"

            result = resolve_model_permission_from_context(context_with_model_permissions, "shared-model")

            assert result.permission.name == "READ"
            assert result.kind == "group-regex"


class TestResolvePromptPermissionFromContext:
    """Tests for resolving prompt permissions from context."""

    @pytest.fixture
    def context_with_prompt_permissions(self):
        """Create context with prompt permissions (using model permissions + prompt regex)."""
        prompt_regex = MagicMock()
        prompt_regex.regex = "^prompt-.*"
        prompt_regex.permission = "EDIT"

        group_prompt_regex = MagicMock()
        group_prompt_regex.regex = "^team-prompt-.*"
        group_prompt_regex.permission = "READ"

        return UserPermissionContext(
            username="testuser",
            group_ids=[1],
            user_experiment_permissions={},
            group_experiment_permissions={},
            experiment_regex_permissions=[],
            group_experiment_regex_permissions=[],
            user_model_permissions={"my-prompt": "MANAGE"},  # Prompts share model permissions
            group_model_permissions={"shared-prompt": "EDIT"},
            model_regex_permissions=[],
            group_model_regex_permissions=[],
            prompt_regex_permissions=[prompt_regex],
            group_prompt_regex_permissions=[group_prompt_regex],
        )

    def test_resolves_user_prompt_permission_from_model_permissions(self, context_with_prompt_permissions):
        """Should resolve user prompt permission from model permissions."""
        with patch("mlflow_oidc_auth.utils.batch_permissions.config") as mock_config:
            mock_config.PERMISSION_SOURCE_ORDER = [
                "user",
                "group",
                "regex",
                "group-regex",
            ]
            mock_config.DEFAULT_MLFLOW_PERMISSION = "NO_PERMISSIONS"

            result = resolve_prompt_permission_from_context(context_with_prompt_permissions, "my-prompt")

            assert result.permission.name == "MANAGE"
            assert result.kind == "user"

    def test_resolves_group_prompt_permission(self, context_with_prompt_permissions):
        """Should resolve group prompt permission from model permissions."""
        with patch("mlflow_oidc_auth.utils.batch_permissions.config") as mock_config:
            mock_config.PERMISSION_SOURCE_ORDER = [
                "user",
                "group",
                "regex",
                "group-regex",
            ]
            mock_config.DEFAULT_MLFLOW_PERMISSION = "NO_PERMISSIONS"

            result = resolve_prompt_permission_from_context(context_with_prompt_permissions, "shared-prompt")

            assert result.permission.name == "EDIT"
            assert result.kind == "group"

    def test_resolves_prompt_regex_permission(self, context_with_prompt_permissions):
        """Should resolve prompt-specific regex permission."""
        with patch("mlflow_oidc_auth.utils.batch_permissions.config") as mock_config:
            mock_config.PERMISSION_SOURCE_ORDER = [
                "user",
                "group",
                "regex",
                "group-regex",
            ]
            mock_config.DEFAULT_MLFLOW_PERMISSION = "NO_PERMISSIONS"

            result = resolve_prompt_permission_from_context(context_with_prompt_permissions, "prompt-test")

            assert result.permission.name == "EDIT"
            assert result.kind == "regex"

    def test_resolves_group_prompt_regex_permission(self, context_with_prompt_permissions):
        """Should resolve group prompt-specific regex permission."""
        with patch("mlflow_oidc_auth.utils.batch_permissions.config") as mock_config:
            mock_config.PERMISSION_SOURCE_ORDER = [
                "user",
                "group",
                "regex",
                "group-regex",
            ]
            mock_config.DEFAULT_MLFLOW_PERMISSION = "NO_PERMISSIONS"

            result = resolve_prompt_permission_from_context(context_with_prompt_permissions, "team-prompt-1")

            assert result.permission.name == "READ"
            assert result.kind == "group-regex"


class TestBatchResolvePermissions:
    """Tests for the batch permission resolution functions."""

    @patch("mlflow_oidc_auth.utils.batch_permissions.build_user_permission_context")
    @patch("mlflow_oidc_auth.utils.batch_permissions.resolve_model_permission_from_context")
    def test_batch_resolve_model_permissions(self, mock_resolve, mock_build_ctx):
        """Should batch resolve permissions for all models."""
        mock_ctx = MagicMock()
        mock_build_ctx.return_value = mock_ctx

        perm_result = MagicMock()
        perm_result.permission.name = "READ"
        perm_result.kind = "user"
        mock_resolve.return_value = perm_result

        model1 = MagicMock()
        model1.name = "model-1"
        model2 = MagicMock()
        model2.name = "model-2"

        result = batch_resolve_model_permissions("testuser", [model1, model2])

        # Should build context once
        mock_build_ctx.assert_called_once_with("testuser")

        # Should resolve for each model
        assert len(result) == 2
        assert "model-1" in result
        assert "model-2" in result
        assert mock_resolve.call_count == 2

    @patch("mlflow_oidc_auth.utils.batch_permissions.build_user_permission_context")
    @patch("mlflow_oidc_auth.utils.batch_permissions.resolve_experiment_permission_from_context")
    def test_batch_resolve_experiment_permissions(self, mock_resolve, mock_build_ctx):
        """Should batch resolve permissions for all experiments."""
        mock_ctx = MagicMock()
        mock_build_ctx.return_value = mock_ctx

        perm_result = MagicMock()
        perm_result.permission.name = "MANAGE"
        perm_result.kind = "group"
        mock_resolve.return_value = perm_result

        exp1 = MagicMock()
        exp1.experiment_id = "1"
        exp1.name = "experiment-1"
        exp2 = MagicMock()
        exp2.experiment_id = "2"
        exp2.name = "experiment-2"

        result = batch_resolve_experiment_permissions("testuser", [exp1, exp2])

        # Should build context once
        mock_build_ctx.assert_called_once_with("testuser")

        # Should resolve for each experiment
        assert len(result) == 2
        assert "1" in result
        assert "2" in result
        assert mock_resolve.call_count == 2

    @patch("mlflow_oidc_auth.utils.batch_permissions.build_user_permission_context")
    @patch("mlflow_oidc_auth.utils.batch_permissions.resolve_prompt_permission_from_context")
    def test_batch_resolve_prompt_permissions(self, mock_resolve, mock_build_ctx):
        """Should batch resolve permissions for all prompts."""
        mock_ctx = MagicMock()
        mock_build_ctx.return_value = mock_ctx

        perm_result = MagicMock()
        perm_result.permission.name = "EDIT"
        perm_result.kind = "regex"
        mock_resolve.return_value = perm_result

        prompt1 = MagicMock()
        prompt1.name = "prompt-1"
        prompt2 = MagicMock()
        prompt2.name = "prompt-2"

        result = batch_resolve_prompt_permissions("testuser", [prompt1, prompt2])

        # Should build context once
        mock_build_ctx.assert_called_once_with("testuser")

        # Should resolve for each prompt
        assert len(result) == 2
        assert "prompt-1" in result
        assert "prompt-2" in result
        assert mock_resolve.call_count == 2

    @patch("mlflow_oidc_auth.utils.batch_permissions.build_user_permission_context")
    def test_batch_resolve_empty_list(self, mock_build_ctx):
        """Should handle empty input lists."""
        mock_ctx = MagicMock()
        mock_build_ctx.return_value = mock_ctx

        result = batch_resolve_model_permissions("testuser", [])

        assert result == {}
        mock_build_ctx.assert_called_once_with("testuser")


class TestFilterManageableFunctions:
    """Tests for the filter_manageable_* helper functions."""

    @patch("mlflow_oidc_auth.utils.batch_permissions.batch_resolve_experiment_permissions")
    def test_filter_manageable_experiments(self, mock_batch_resolve):
        """Should filter experiments to only those user can manage."""
        exp1 = MagicMock()
        exp1.experiment_id = "1"
        exp1.name = "can-manage"

        exp2 = MagicMock()
        exp2.experiment_id = "2"
        exp2.name = "cannot-manage"

        exp3 = MagicMock()
        exp3.experiment_id = "3"
        exp3.name = "also-can-manage"

        perm_manage = MagicMock()
        perm_manage.permission.can_manage = True

        perm_no_manage = MagicMock()
        perm_no_manage.permission.can_manage = False

        mock_batch_resolve.return_value = {
            "1": perm_manage,
            "2": perm_no_manage,
            "3": perm_manage,
        }

        result = filter_manageable_experiments("testuser", [exp1, exp2, exp3])

        assert len(result) == 2
        assert exp1 in result
        assert exp2 not in result
        assert exp3 in result
        mock_batch_resolve.assert_called_once_with("testuser", [exp1, exp2, exp3])

    @patch("mlflow_oidc_auth.utils.batch_permissions.batch_resolve_model_permissions")
    def test_filter_manageable_models(self, mock_batch_resolve):
        """Should filter models to only those user can manage."""
        model1 = MagicMock()
        model1.name = "can-manage"

        model2 = MagicMock()
        model2.name = "cannot-manage"

        perm_manage = MagicMock()
        perm_manage.permission.can_manage = True

        perm_no_manage = MagicMock()
        perm_no_manage.permission.can_manage = False

        mock_batch_resolve.return_value = {
            "can-manage": perm_manage,
            "cannot-manage": perm_no_manage,
        }

        result = filter_manageable_models("testuser", [model1, model2])

        assert len(result) == 1
        assert model1 in result
        assert model2 not in result
        mock_batch_resolve.assert_called_once_with("testuser", [model1, model2])

    @patch("mlflow_oidc_auth.utils.batch_permissions.batch_resolve_prompt_permissions")
    def test_filter_manageable_prompts(self, mock_batch_resolve):
        """Should filter prompts to only those user can manage."""
        prompt1 = MagicMock()
        prompt1.name = "can-manage"

        prompt2 = MagicMock()
        prompt2.name = "cannot-manage"

        perm_manage = MagicMock()
        perm_manage.permission.can_manage = True

        perm_no_manage = MagicMock()
        perm_no_manage.permission.can_manage = False

        mock_batch_resolve.return_value = {
            "can-manage": perm_manage,
            "cannot-manage": perm_no_manage,
        }

        result = filter_manageable_prompts("testuser", [prompt1, prompt2])

        assert len(result) == 1
        assert prompt1 in result
        assert prompt2 not in result
        mock_batch_resolve.assert_called_once_with("testuser", [prompt1, prompt2])

    @patch("mlflow_oidc_auth.utils.batch_permissions.batch_resolve_experiment_permissions")
    def test_filter_manageable_returns_empty_when_none_manageable(self, mock_batch_resolve):
        """Should return empty list when no items are manageable."""
        exp1 = MagicMock()
        exp1.experiment_id = "1"

        perm_no_manage = MagicMock()
        perm_no_manage.permission.can_manage = False

        mock_batch_resolve.return_value = {"1": perm_no_manage}

        result = filter_manageable_experiments("testuser", [exp1])

        assert result == []

    @patch("mlflow_oidc_auth.utils.batch_permissions.batch_resolve_model_permissions")
    def test_filter_manageable_returns_all_when_all_manageable(self, mock_batch_resolve):
        """Should return all items when all are manageable."""
        model1 = MagicMock()
        model1.name = "model-1"

        model2 = MagicMock()
        model2.name = "model-2"

        perm_manage = MagicMock()
        perm_manage.permission.can_manage = True

        mock_batch_resolve.return_value = {
            "model-1": perm_manage,
            "model-2": perm_manage,
        }

        result = filter_manageable_models("testuser", [model1, model2])

        assert len(result) == 2
        assert model1 in result
        assert model2 in result

    @patch("mlflow_oidc_auth.utils.batch_permissions.batch_resolve_prompt_permissions")
    def test_filter_manageable_empty_input(self, mock_batch_resolve):
        """Should handle empty input list."""
        mock_batch_resolve.return_value = {}

        result = filter_manageable_prompts("testuser", [])

        assert result == []



"""Tests for gateway CRUD endpoints in user_permissions router.

This module tests all 30 gateway-related endpoints in the user_permissions router:
- 5 for gateway endpoint permissions (list, create, get, update, delete)
- 5 for gateway endpoint pattern permissions
- 5 for gateway model definition permissions
- 5 for gateway model definition pattern permissions
- 5 for gateway secret permissions
- 5 for gateway secret pattern permissions
"""

from unittest.mock import MagicMock, patch

import pytest

from mlflow_oidc_auth.dependencies import check_admin_permission
from mlflow_oidc_auth.utils import get_is_admin, get_username

USER_BASE = "/api/2.0/mlflow/permissions/users"
_UP = "mlflow_oidc_auth.routers.user_permissions"


def _make_perm_result(permission_name: str, kind: str = "user") -> MagicMock:
    """Create a mock PermissionResult for effective_gateway_*_permission."""
    result = MagicMock()
    result.permission.name = permission_name
    result.permission.can_manage = permission_name == "MANAGE"
    result.kind = kind
    return result


@pytest.fixture
def override_admin(test_app):
    """Override admin permission check to always pass."""

    async def always_admin():
        return "admin@example.com"

    test_app.dependency_overrides[check_admin_permission] = always_admin
    yield
    test_app.dependency_overrides.pop(check_admin_permission, None)


@pytest.fixture
def override_user_non_admin(test_app):
    """Override get_username and get_is_admin for non-admin user tests."""

    async def _username():
        return "test@example.com"

    async def _is_admin():
        return False

    test_app.dependency_overrides[get_username] = _username
    test_app.dependency_overrides[get_is_admin] = _is_admin
    yield
    test_app.dependency_overrides.pop(get_username, None)
    test_app.dependency_overrides.pop(get_is_admin, None)


def _make_direct_perm(attr_name: str, value: str, permission: str = "READ") -> MagicMock:
    """Create a mock direct permission entity."""
    perm = MagicMock()
    setattr(perm, attr_name, value)
    perm.permission = permission
    return perm


def _make_regex_perm(
    perm_id: int = 1,
    regex: str = ".*",
    priority: int = 1,
    user_id: int = 5,
    permission: str = "READ",
) -> MagicMock:
    """Create a mock regex permission entity."""
    perm = MagicMock()
    perm.id = perm_id
    perm.regex = regex
    perm.priority = priority
    perm.user_id = user_id
    perm.permission = permission
    return perm


# ========================================================================================
# GATEWAY ENDPOINT PERMISSIONS
# ========================================================================================


@pytest.mark.usefixtures("authenticated_session", "override_admin")
class TestUserGatewayEndpointPermissions:
    """Tests for user gateway endpoint CRUD endpoints (create/get/update/delete)."""

    def test_create(self, authenticated_client, mock_store):
        """Test creating a user gateway endpoint permission."""
        mock_store.create_gateway_endpoint_permission.return_value = _make_direct_perm("endpoint_id", "ep-1", "MANAGE")
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.post(
                f"{USER_BASE}/user@example.com/gateways/endpoints/ep-1",
                json={"permission": "MANAGE"},
            )
        assert resp.status_code == 201
        assert resp.json()["name"] == "ep-1"
        assert resp.json()["kind"] == "user"

    def test_get(self, authenticated_client, mock_store):
        """Test getting a specific user gateway endpoint permission."""
        mock_store.get_gateway_endpoint_permission.return_value = _make_direct_perm("endpoint_id", "ep-1", "READ")
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.get(f"{USER_BASE}/user@example.com/gateways/endpoints/ep-1")
        assert resp.status_code == 200
        assert resp.json()["name"] == "ep-1"

    def test_update(self, authenticated_client, mock_store):
        """Test updating a user gateway endpoint permission."""
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.patch(
                f"{USER_BASE}/user@example.com/gateways/endpoints/ep-1",
                json={"permission": "EDIT"},
            )
        assert resp.status_code == 200
        assert "updated" in resp.json()["message"].lower()
        mock_store.update_gateway_endpoint_permission.assert_called_once()

    def test_delete(self, authenticated_client, mock_store):
        """Test deleting a user gateway endpoint permission."""
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.delete(f"{USER_BASE}/user@example.com/gateways/endpoints/ep-1")
        assert resp.status_code == 200
        assert "deleted" in resp.json()["message"].lower()
        mock_store.delete_gateway_endpoint_permission.assert_called_once()

    def test_get_not_found(self, authenticated_client, mock_store):
        """Test 404 for missing permission."""
        mock_store.get_gateway_endpoint_permission.side_effect = Exception("Not found")
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.get(f"{USER_BASE}/user@example.com/gateways/endpoints/missing")
        assert resp.status_code == 404


# ========================================================================================
# GATEWAY ENDPOINT PATTERN PERMISSIONS
# ========================================================================================


@pytest.mark.usefixtures("authenticated_session", "override_admin")
class TestUserGatewayEndpointPatternPermissions:
    """Tests for user gateway endpoint pattern CRUD endpoints."""

    def test_list(self, authenticated_client, mock_store):
        """Test listing user gateway endpoint regex permissions."""
        mock_store.list_gateway_endpoint_regex_permissions.return_value = [
            _make_regex_perm(1, "ep-.*", 1, 5, "READ"),
        ]
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.get(f"{USER_BASE}/user@example.com/gateways/endpoints-patterns")
        assert resp.status_code == 200
        assert resp.json()[0]["regex"] == "ep-.*"

    def test_create(self, authenticated_client, mock_store):
        """Test creating a user gateway endpoint regex permission."""
        mock_store.create_gateway_endpoint_regex_permission.return_value = _make_regex_perm(1, "ep-.*", 1, 5, "READ")
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.post(
                f"{USER_BASE}/user@example.com/gateways/endpoints-patterns",
                json={"regex": "ep-.*", "priority": 1, "permission": "READ"},
            )
        assert resp.status_code == 201

    def test_get(self, authenticated_client, mock_store):
        """Test getting a specific user gateway endpoint regex permission."""
        mock_store.get_gateway_endpoint_regex_permission.return_value = _make_regex_perm(1, "ep-.*", 1, 5, "READ")
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.get(f"{USER_BASE}/user@example.com/gateways/endpoints-patterns/1")
        assert resp.status_code == 200

    def test_update(self, authenticated_client, mock_store):
        """Test updating a user gateway endpoint regex permission."""
        mock_store.update_gateway_endpoint_regex_permission.return_value = _make_regex_perm(1, "new-.*", 2, 5, "MANAGE")
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.patch(
                f"{USER_BASE}/user@example.com/gateways/endpoints-patterns/1",
                json={"regex": "new-.*", "priority": 2, "permission": "MANAGE"},
            )
        assert resp.status_code == 200

    def test_delete(self, authenticated_client, mock_store):
        """Test deleting a user gateway endpoint regex permission."""
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.delete(f"{USER_BASE}/user@example.com/gateways/endpoints-patterns/1")
        assert resp.status_code == 200
        mock_store.delete_gateway_endpoint_regex_permission.assert_called_once()

    def test_get_not_found(self, authenticated_client, mock_store):
        """Test 404 for missing regex permission."""
        mock_store.get_gateway_endpoint_regex_permission.side_effect = Exception("Not found")
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.get(f"{USER_BASE}/user@example.com/gateways/endpoints-patterns/999")
        assert resp.status_code == 404


# ========================================================================================
# GATEWAY MODEL DEFINITION PERMISSIONS
# ========================================================================================


@pytest.mark.usefixtures("authenticated_session", "override_admin")
class TestUserGatewayModelDefinitionPermissions:
    """Tests for user gateway model definition CRUD endpoints (create/get/update/delete)."""

    def test_create(self, authenticated_client, mock_store):
        """Test creating a user gateway model definition permission."""
        mock_store.create_gateway_model_definition_permission.return_value = _make_direct_perm("model_definition_id", "gpt-4", "MANAGE")
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.post(
                f"{USER_BASE}/user@example.com/gateways/model-definitions/gpt-4",
                json={"permission": "MANAGE"},
            )
        assert resp.status_code == 201
        assert resp.json()["name"] == "gpt-4"

    def test_get(self, authenticated_client, mock_store):
        """Test getting a specific user gateway model definition permission."""
        mock_store.get_gateway_model_definition_permission.return_value = _make_direct_perm("model_definition_id", "gpt-4", "READ")
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.get(f"{USER_BASE}/user@example.com/gateways/model-definitions/gpt-4")
        assert resp.status_code == 200

    def test_update(self, authenticated_client, mock_store):
        """Test updating a user gateway model definition permission."""
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.patch(
                f"{USER_BASE}/user@example.com/gateways/model-definitions/gpt-4",
                json={"permission": "EDIT"},
            )
        assert resp.status_code == 200
        mock_store.update_gateway_model_definition_permission.assert_called_once()

    def test_delete(self, authenticated_client, mock_store):
        """Test deleting a user gateway model definition permission."""
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.delete(f"{USER_BASE}/user@example.com/gateways/model-definitions/gpt-4")
        assert resp.status_code == 200
        mock_store.delete_gateway_model_definition_permission.assert_called_once()

    def test_get_not_found(self, authenticated_client, mock_store):
        """Test 404 for missing permission."""
        mock_store.get_gateway_model_definition_permission.side_effect = Exception("Not found")
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.get(f"{USER_BASE}/user@example.com/gateways/model-definitions/missing")
        assert resp.status_code == 404


# ========================================================================================
# GATEWAY MODEL DEFINITION PATTERN PERMISSIONS
# ========================================================================================


@pytest.mark.usefixtures("authenticated_session", "override_admin")
class TestUserGatewayModelDefinitionPatternPermissions:
    """Tests for user gateway model definition pattern CRUD endpoints."""

    def test_list(self, authenticated_client, mock_store):
        """Test listing user gateway model definition regex permissions."""
        mock_store.list_gateway_model_definition_regex_permissions.return_value = [
            _make_regex_perm(1, "gpt-.*", 1, 5, "READ"),
        ]
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.get(f"{USER_BASE}/user@example.com/gateways/model-definitions-patterns")
        assert resp.status_code == 200
        assert resp.json()[0]["regex"] == "gpt-.*"

    def test_create(self, authenticated_client, mock_store):
        """Test creating a user gateway model definition regex permission."""
        mock_store.create_gateway_model_definition_regex_permission.return_value = _make_regex_perm(2, "gpt-.*", 1, 5, "READ")
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.post(
                f"{USER_BASE}/user@example.com/gateways/model-definitions-patterns",
                json={"regex": "gpt-.*", "priority": 1, "permission": "READ"},
            )
        assert resp.status_code == 201

    def test_get(self, authenticated_client, mock_store):
        """Test getting a specific user gateway model definition regex permission."""
        mock_store.get_gateway_model_definition_regex_permission.return_value = _make_regex_perm(1, "gpt-.*", 1, 5, "READ")
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.get(f"{USER_BASE}/user@example.com/gateways/model-definitions-patterns/1")
        assert resp.status_code == 200

    def test_update(self, authenticated_client, mock_store):
        """Test updating a user gateway model definition regex permission."""
        mock_store.update_gateway_model_definition_regex_permission.return_value = _make_regex_perm(1, "claude-.*", 2, 5, "MANAGE")
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.patch(
                f"{USER_BASE}/user@example.com/gateways/model-definitions-patterns/1",
                json={"regex": "claude-.*", "priority": 2, "permission": "MANAGE"},
            )
        assert resp.status_code == 200

    def test_delete(self, authenticated_client, mock_store):
        """Test deleting a user gateway model definition regex permission."""
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.delete(f"{USER_BASE}/user@example.com/gateways/model-definitions-patterns/1")
        assert resp.status_code == 200
        mock_store.delete_gateway_model_definition_regex_permission.assert_called_once()

    def test_get_not_found(self, authenticated_client, mock_store):
        """Test 404 for missing regex permission."""
        mock_store.get_gateway_model_definition_regex_permission.side_effect = Exception("Not found")
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.get(f"{USER_BASE}/user@example.com/gateways/model-definitions-patterns/999")
        assert resp.status_code == 404


# ========================================================================================
# GATEWAY SECRET PERMISSIONS
# ========================================================================================


@pytest.mark.usefixtures("authenticated_session", "override_admin")
class TestUserGatewaySecretPermissions:
    """Tests for user gateway secret CRUD endpoints (create/get/update/delete)."""

    def test_create(self, authenticated_client, mock_store):
        """Test creating a user gateway secret permission."""
        mock_store.create_gateway_secret_permission.return_value = _make_direct_perm("secret_id", "api-key", "MANAGE")
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.post(
                f"{USER_BASE}/user@example.com/gateways/secrets/api-key",
                json={"permission": "MANAGE"},
            )
        assert resp.status_code == 201
        assert resp.json()["name"] == "api-key"

    def test_get(self, authenticated_client, mock_store):
        """Test getting a specific user gateway secret permission."""
        mock_store.get_gateway_secret_permission.return_value = _make_direct_perm("secret_id", "api-key", "READ")
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.get(f"{USER_BASE}/user@example.com/gateways/secrets/api-key")
        assert resp.status_code == 200

    def test_update(self, authenticated_client, mock_store):
        """Test updating a user gateway secret permission."""
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.patch(
                f"{USER_BASE}/user@example.com/gateways/secrets/api-key",
                json={"permission": "EDIT"},
            )
        assert resp.status_code == 200
        mock_store.update_gateway_secret_permission.assert_called_once()

    def test_delete(self, authenticated_client, mock_store):
        """Test deleting a user gateway secret permission."""
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.delete(f"{USER_BASE}/user@example.com/gateways/secrets/api-key")
        assert resp.status_code == 200
        mock_store.delete_gateway_secret_permission.assert_called_once()

    def test_get_not_found(self, authenticated_client, mock_store):
        """Test 404 for missing permission."""
        mock_store.get_gateway_secret_permission.side_effect = Exception("Not found")
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.get(f"{USER_BASE}/user@example.com/gateways/secrets/missing")
        assert resp.status_code == 404


# ========================================================================================
# GATEWAY SECRET PATTERN PERMISSIONS
# ========================================================================================


@pytest.mark.usefixtures("authenticated_session", "override_admin")
class TestUserGatewaySecretPatternPermissions:
    """Tests for user gateway secret pattern CRUD endpoints."""

    def test_list(self, authenticated_client, mock_store):
        """Test listing user gateway secret regex permissions."""
        mock_store.list_gateway_secret_regex_permissions.return_value = [
            _make_regex_perm(1, "api-.*", 1, 5, "READ"),
        ]
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.get(f"{USER_BASE}/user@example.com/gateways/secrets-patterns")
        assert resp.status_code == 200
        assert resp.json()[0]["regex"] == "api-.*"

    def test_create(self, authenticated_client, mock_store):
        """Test creating a user gateway secret regex permission."""
        mock_store.create_gateway_secret_regex_permission.return_value = _make_regex_perm(2, "api-.*", 1, 5, "READ")
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.post(
                f"{USER_BASE}/user@example.com/gateways/secrets-patterns",
                json={"regex": "api-.*", "priority": 1, "permission": "READ"},
            )
        assert resp.status_code == 201

    def test_get(self, authenticated_client, mock_store):
        """Test getting a specific user gateway secret regex permission."""
        mock_store.get_gateway_secret_regex_permission.return_value = _make_regex_perm(1, "api-.*", 1, 5, "READ")
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.get(f"{USER_BASE}/user@example.com/gateways/secrets-patterns/1")
        assert resp.status_code == 200

    def test_update(self, authenticated_client, mock_store):
        """Test updating a user gateway secret regex permission."""
        mock_store.update_gateway_secret_regex_permission.return_value = _make_regex_perm(1, "key-.*", 2, 5, "MANAGE")
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.patch(
                f"{USER_BASE}/user@example.com/gateways/secrets-patterns/1",
                json={"regex": "key-.*", "priority": 2, "permission": "MANAGE"},
            )
        assert resp.status_code == 200

    def test_delete(self, authenticated_client, mock_store):
        """Test deleting a user gateway secret regex permission."""
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.delete(f"{USER_BASE}/user@example.com/gateways/secrets-patterns/1")
        assert resp.status_code == 200
        mock_store.delete_gateway_secret_regex_permission.assert_called_once()

    def test_get_not_found(self, authenticated_client, mock_store):
        """Test 404 for missing regex permission."""
        mock_store.get_gateway_secret_regex_permission.side_effect = Exception("Not found")
        with patch("mlflow_oidc_auth.routers.user_permissions.store", mock_store):
            resp = authenticated_client.get(f"{USER_BASE}/user@example.com/gateways/secrets-patterns/999")
        assert resp.status_code == 404


# ========================================================================================
# GATEWAY LIST ENDPOINTS (non-admin access)
# ========================================================================================


@pytest.mark.usefixtures("authenticated_session")
class TestUserGatewayEndpointPermissionsList:
    """Tests for listing gateway endpoint permissions with role-based filtering."""

    def test_list_as_admin(self, admin_client):
        """Admin sees all gateway endpoints with target user's effective permissions."""
        endpoints = [{"name": "ep-1"}, {"name": "ep-2"}]
        with (
            patch(f"{_UP}.fetch_all_gateway_endpoints", return_value=endpoints),
            patch(
                f"{_UP}.effective_gateway_endpoint_permission",
                side_effect=[
                    _make_perm_result("MANAGE", "user"),
                    _make_perm_result("READ", "regex"),
                ],
            ),
        ):
            resp = admin_client.get(f"{USER_BASE}/user@example.com/gateways/endpoints")
        assert resp.status_code == 200
        body = resp.json()
        assert len(body) == 2
        assert {"name": "ep-1", "permission": "MANAGE", "kind": "user"} in body
        assert {"name": "ep-2", "permission": "READ", "kind": "regex"} in body

    @pytest.mark.usefixtures("override_user_non_admin")
    def test_list_same_user_filters_no_permissions(self, authenticated_client):
        """Same user sees endpoints where permission != NO_PERMISSIONS."""
        endpoints = [{"name": "ep-visible"}, {"name": "ep-hidden"}]
        with (
            patch(f"{_UP}.fetch_all_gateway_endpoints", return_value=endpoints),
            patch(
                f"{_UP}.effective_gateway_endpoint_permission",
                side_effect=[
                    _make_perm_result("READ", "user"),
                    _make_perm_result("NO_PERMISSIONS", "default"),
                ],
            ),
        ):
            resp = authenticated_client.get(f"{USER_BASE}/test@example.com/gateways/endpoints")
        assert resp.status_code == 200
        body = resp.json()
        assert len(body) == 1
        assert body[0]["name"] == "ep-visible"

    @pytest.mark.usefixtures("override_user_non_admin")
    def test_list_other_user_shows_only_manageable(self, authenticated_client):
        """Non-admin querying another user sees only endpoints they can manage."""
        endpoints = [{"name": "ep-manage"}, {"name": "ep-readonly"}]
        with (
            patch(f"{_UP}.fetch_all_gateway_endpoints", return_value=endpoints),
            patch(
                f"{_UP}.effective_gateway_endpoint_permission",
                side_effect=[
                    # target user's permission for ep-manage
                    _make_perm_result("READ", "user"),
                    # caller's permission for ep-manage (can_manage=True)
                    _make_perm_result("MANAGE", "user"),
                    # target user's permission for ep-readonly
                    _make_perm_result("READ", "user"),
                    # caller's permission for ep-readonly (can_manage=False)
                    _make_perm_result("READ", "regex"),
                ],
            ),
        ):
            resp = authenticated_client.get(f"{USER_BASE}/other@example.com/gateways/endpoints")
        assert resp.status_code == 200
        body = resp.json()
        assert len(body) == 1
        assert body[0]["name"] == "ep-manage"

    def test_list_error(self, admin_client):
        """Test error handling for list endpoint."""
        with patch(f"{_UP}.fetch_all_gateway_endpoints", side_effect=Exception("DB error")):
            resp = admin_client.get(f"{USER_BASE}/user@example.com/gateways/endpoints")
        assert resp.status_code == 500


@pytest.mark.usefixtures("authenticated_session")
class TestUserGatewayModelDefinitionPermissionsList:
    """Tests for listing gateway model definition permissions with role-based filtering."""

    def test_list_as_admin(self, admin_client):
        """Admin sees all gateway model definitions with target user's effective permissions."""
        models = [{"name": "gpt-4"}, {"name": "claude-3"}]
        with (
            patch(f"{_UP}.fetch_all_gateway_model_definitions", return_value=models),
            patch(
                f"{_UP}.effective_gateway_model_definition_permission",
                side_effect=[
                    _make_perm_result("MANAGE", "user"),
                    _make_perm_result("READ", "regex"),
                ],
            ),
        ):
            resp = admin_client.get(f"{USER_BASE}/user@example.com/gateways/model-definitions")
        assert resp.status_code == 200
        body = resp.json()
        assert len(body) == 2
        assert {"name": "gpt-4", "permission": "MANAGE", "kind": "user"} in body

    @pytest.mark.usefixtures("override_user_non_admin")
    def test_list_same_user_filters_no_permissions(self, authenticated_client):
        """Same user sees model definitions where permission != NO_PERMISSIONS."""
        models = [{"name": "gpt-4"}, {"name": "hidden-model"}]
        with (
            patch(f"{_UP}.fetch_all_gateway_model_definitions", return_value=models),
            patch(
                f"{_UP}.effective_gateway_model_definition_permission",
                side_effect=[
                    _make_perm_result("READ", "user"),
                    _make_perm_result("NO_PERMISSIONS", "default"),
                ],
            ),
        ):
            resp = authenticated_client.get(f"{USER_BASE}/test@example.com/gateways/model-definitions")
        assert resp.status_code == 200
        body = resp.json()
        assert len(body) == 1
        assert body[0]["name"] == "gpt-4"

    @pytest.mark.usefixtures("override_user_non_admin")
    def test_list_other_user_shows_only_manageable(self, authenticated_client):
        """Non-admin querying another user sees only model definitions they can manage."""
        models = [{"name": "gpt-4"}, {"name": "claude-3"}]
        with (
            patch(f"{_UP}.fetch_all_gateway_model_definitions", return_value=models),
            patch(
                f"{_UP}.effective_gateway_model_definition_permission",
                side_effect=[
                    _make_perm_result("READ", "user"),
                    _make_perm_result("MANAGE", "user"),
                    _make_perm_result("READ", "user"),
                    _make_perm_result("READ", "regex"),
                ],
            ),
        ):
            resp = authenticated_client.get(f"{USER_BASE}/other@example.com/gateways/model-definitions")
        assert resp.status_code == 200
        body = resp.json()
        assert len(body) == 1
        assert body[0]["name"] == "gpt-4"

    def test_list_error(self, admin_client):
        """Test error handling for list endpoint."""
        with patch(f"{_UP}.fetch_all_gateway_model_definitions", side_effect=Exception("DB error")):
            resp = admin_client.get(f"{USER_BASE}/user@example.com/gateways/model-definitions")
        assert resp.status_code == 500


@pytest.mark.usefixtures("authenticated_session")
class TestUserGatewaySecretPermissionsList:
    """Tests for listing gateway secret permissions with role-based filtering."""

    def test_list_as_admin(self, admin_client):
        """Admin sees all gateway secrets with target user's effective permissions."""
        secrets = [{"secret_name": "api-key"}, {"secret_name": "db-pass"}]
        with (
            patch(f"{_UP}.fetch_all_gateway_secrets", return_value=secrets),
            patch(
                f"{_UP}.effective_gateway_secret_permission",
                side_effect=[
                    _make_perm_result("MANAGE", "user"),
                    _make_perm_result("READ", "regex"),
                ],
            ),
        ):
            resp = admin_client.get(f"{USER_BASE}/user@example.com/gateways/secrets")
        assert resp.status_code == 200
        body = resp.json()
        assert len(body) == 2
        assert {"name": "api-key", "permission": "MANAGE", "kind": "user"} in body
        assert {"name": "db-pass", "permission": "READ", "kind": "regex"} in body

    @pytest.mark.usefixtures("override_user_non_admin")
    def test_list_same_user_filters_no_permissions(self, authenticated_client):
        """Same user sees secrets where permission != NO_PERMISSIONS."""
        secrets = [{"secret_name": "visible-key"}, {"secret_name": "hidden-key"}]
        with (
            patch(f"{_UP}.fetch_all_gateway_secrets", return_value=secrets),
            patch(
                f"{_UP}.effective_gateway_secret_permission",
                side_effect=[
                    _make_perm_result("READ", "user"),
                    _make_perm_result("NO_PERMISSIONS", "default"),
                ],
            ),
        ):
            resp = authenticated_client.get(f"{USER_BASE}/test@example.com/gateways/secrets")
        assert resp.status_code == 200
        body = resp.json()
        assert len(body) == 1
        assert body[0]["name"] == "visible-key"

    @pytest.mark.usefixtures("override_user_non_admin")
    def test_list_other_user_shows_only_manageable(self, authenticated_client):
        """Non-admin querying another user sees only secrets they can manage."""
        secrets = [{"secret_name": "manageable"}, {"secret_name": "readonly"}]
        with (
            patch(f"{_UP}.fetch_all_gateway_secrets", return_value=secrets),
            patch(
                f"{_UP}.effective_gateway_secret_permission",
                side_effect=[
                    _make_perm_result("READ", "user"),
                    _make_perm_result("MANAGE", "user"),
                    _make_perm_result("READ", "user"),
                    _make_perm_result("READ", "regex"),
                ],
            ),
        ):
            resp = authenticated_client.get(f"{USER_BASE}/other@example.com/gateways/secrets")
        assert resp.status_code == 200
        body = resp.json()
        assert len(body) == 1
        assert body[0]["name"] == "manageable"

    def test_list_error(self, admin_client):
        """Test error handling for list endpoint."""
        with patch(f"{_UP}.fetch_all_gateway_secrets", side_effect=Exception("DB error")):
            resp = admin_client.get(f"{USER_BASE}/user@example.com/gateways/secrets")
        assert resp.status_code == 500



import unittest

from mlflow_oidc_auth.entities import (
    GatewayModelDefinitionPermission,
    GatewayModelDefinitionRegexPermission,
    GatewayModelDefinitionGroupRegexPermission,
)


class TestGatewayModelDefinitionEntities(unittest.TestCase):
    def test_model_definition_permission_properties(self):
        perm = GatewayModelDefinitionPermission(model_definition_id="md1", permission="USE", user_id=1)
        self.assertEqual(perm.model_definition_id, "md1")
        self.assertEqual(perm.permission, "USE")
        self.assertEqual(perm.user_id, 1)

    def test_model_definition_permission_to_json_and_from_json(self):
        perm = GatewayModelDefinitionPermission(model_definition_id="md1", permission="READ", user_id=1, group_id=2)
        json_data = perm.to_json()
        self.assertEqual(json_data["model_definition_id"], "md1")
        self.assertEqual(json_data["permission"], "READ")
        self.assertEqual(json_data["user_id"], 1)
        self.assertEqual(json_data["group_id"], 2)

        p2 = GatewayModelDefinitionPermission.from_json(json_data)
        self.assertEqual(p2.model_definition_id, "md1")
        self.assertEqual(p2.permission, "READ")
        self.assertEqual(p2.user_id, 1)
        self.assertEqual(p2.group_id, 2)

    def test_model_definition_regex_permission(self):
        perm = GatewayModelDefinitionRegexPermission(id_="1", regex="md-.*", priority=5, user_id=1, permission="USE")
        self.assertEqual(perm.id, "1")
        self.assertEqual(perm.regex, "md-.*")
        self.assertEqual(perm.priority, 5)
        self.assertEqual(perm.user_id, 1)
        self.assertEqual(perm.permission, "USE")

        json_data = perm.to_json()
        expected = {
            "id": "1",
            "regex": "md-.*",
            "priority": 5,
            "user_id": 1,
            "permission": "USE",
        }
        self.assertEqual(json_data, expected)

        p2 = GatewayModelDefinitionRegexPermission.from_json(json_data)
        self.assertEqual(p2.id, "1")
        self.assertEqual(p2.user_id, 1)
        self.assertEqual(p2.permission, "USE")

    def test_model_definition_group_regex_permission(self):
        perm = GatewayModelDefinitionGroupRegexPermission(id_="1", regex="md-.*", priority=5, group_id=1, permission="READ")
        self.assertEqual(perm.id, "1")
        self.assertEqual(perm.regex, "md-.*")
        self.assertEqual(perm.priority, 5)
        self.assertEqual(perm.group_id, 1)
        self.assertEqual(perm.permission, "READ")

        json_data = perm.to_json()
        expected = {
            "id": "1",
            "regex": "md-.*",
            "priority": 5,
            "group_id": 1,
            "permission": "READ",
        }
        self.assertEqual(json_data, expected)

        p2 = GatewayModelDefinitionGroupRegexPermission.from_json(json_data)
        self.assertEqual(p2.id, "1")
        self.assertEqual(p2.group_id, 1)
        self.assertEqual(p2.permission, "READ")



"""Validators for AI Gateway resources (endpoints, secrets, model definitions).

These validators enforce permission checks for gateway CRUD operations
in the Flask before-request hook.
"""

from __future__ import annotations

from flask import request

from mlflow_oidc_auth.logger import get_logger
from mlflow_oidc_auth.utils import get_request_param
from mlflow_oidc_auth.utils.permissions import (
    can_manage_gateway_endpoint,
    can_manage_gateway_model_definition,
    can_manage_gateway_secret,
    can_read_gateway_endpoint,
    can_read_gateway_model_definition,
    can_read_gateway_secret,
    can_update_gateway_endpoint,
    can_update_gateway_model_definition,
    can_update_gateway_secret,
)

_logger = get_logger()


# ---------------------------------------------------------------------------
# Gateway Endpoint validators
# ---------------------------------------------------------------------------


def validate_can_read_gateway_endpoint(username: str) -> bool:
    """Validate READ permission on a gateway endpoint.

    ``GetGatewayEndpoint`` provides ``name`` (or ``endpoint_id``).
    We check by name when available, falling back to ID resolution.
    """
    name = _get_gateway_endpoint_name()
    if not name:
        return False
    return can_read_gateway_endpoint(name, username)


def validate_can_update_gateway_endpoint(username: str) -> bool:
    """Validate UPDATE permission on a gateway endpoint.

    The old name is already stashed in ``flask.g`` by
    ``_stash_gateway_context`` (which runs for all users).
    For the permission check we resolve the current endpoint name
    via ``endpoint_id`` since the request ``name`` is the *new* name.
    """
    name = _get_gateway_endpoint_name_for_update()
    if not name:
        return False
    return can_update_gateway_endpoint(name, username)


def validate_can_delete_gateway_endpoint(username: str) -> bool:
    """Validate MANAGE permission for deleting a gateway endpoint.

    The name is already stashed in ``flask.g`` by
    ``_stash_gateway_context`` (which runs for all users).
    """
    name = _get_gateway_endpoint_name()
    if not name:
        return False
    return can_manage_gateway_endpoint(name, username)


def validate_can_manage_gateway_endpoint_validator(username: str) -> bool:
    """Validate MANAGE permission on a gateway endpoint."""
    name = _get_gateway_endpoint_name()
    if not name:
        return False
    return can_manage_gateway_endpoint(name, username)


# ---------------------------------------------------------------------------
# Gateway Secret validators
# ---------------------------------------------------------------------------


def validate_can_read_gateway_secret(username: str) -> bool:
    """Validate READ permission on a gateway secret.

    ``GetGatewaySecretInfo`` provides ``secret_name`` (or ``secret_id``).
    """
    name = _get_gateway_secret_name()
    if not name:
        return False
    return can_read_gateway_secret(name, username)


def validate_can_update_gateway_secret(username: str) -> bool:
    """Validate UPDATE permission on a gateway secret."""
    name = _get_gateway_secret_name()
    if not name:
        return False
    return can_update_gateway_secret(name, username)


def validate_can_delete_gateway_secret(username: str) -> bool:
    """Validate MANAGE permission for deleting a gateway secret.

    The name is already stashed in ``flask.g`` by
    ``_stash_gateway_context`` (which runs for all users).
    """
    name = _get_gateway_secret_name()
    if not name:
        return False
    return can_manage_gateway_secret(name, username)


# ---------------------------------------------------------------------------
# Gateway Model Definition validators
# ---------------------------------------------------------------------------


def validate_can_read_gateway_model_definition(username: str) -> bool:
    """Validate READ permission on a gateway model definition.

    ``GetGatewayModelDefinition`` only provides ``model_definition_id``,
    so we fall back to ID resolution via the tracking store. If
    resolution fails, we allow through and rely on list filtering.
    """
    name = _get_gateway_model_definition_name()
    if not name:
        # ID-only request — cannot resolve name; allow and rely on list filtering
        return True
    return can_read_gateway_model_definition(name, username)


def validate_can_update_gateway_model_definition(username: str) -> bool:
    """Validate UPDATE permission on a gateway model definition."""
    name = _get_gateway_model_definition_name()
    if not name:
        return True
    return can_update_gateway_model_definition(name, username)


def validate_can_delete_gateway_model_definition(username: str) -> bool:
    """Validate MANAGE permission for deleting a gateway model definition.

    The name is already stashed in ``flask.g`` by
    ``_stash_gateway_context`` (which runs for all users).
    """
    name = _get_gateway_model_definition_name()
    if not name:
        return True
    return can_manage_gateway_model_definition(name, username)


# ---------------------------------------------------------------------------
# Internal helpers
# ---------------------------------------------------------------------------


def _resolve_endpoint_name_from_id(endpoint_id: str) -> str | None:
    """Look up a gateway endpoint's name from its ID via the tracking store."""
    try:
        from mlflow.server.handlers import _get_tracking_store

        endpoint = _get_tracking_store().get_gateway_endpoint(endpoint_id=endpoint_id)
        return endpoint.name
    except Exception:
        _logger.debug(f"Could not resolve gateway endpoint name from id '{endpoint_id}'")
        return None


def _resolve_secret_name_from_id(secret_id: str) -> str | None:
    """Look up a gateway secret's name from its ID via the tracking store."""
    try:
        from mlflow.server.handlers import _get_tracking_store

        secret = _get_tracking_store().get_secret_info(secret_id=secret_id)
        return secret.secret_name
    except Exception:
        _logger.debug(f"Could not resolve gateway secret name from id '{secret_id}'")
        return None


def _resolve_model_definition_name_from_id(model_definition_id: str) -> str | None:
    """Look up a gateway model definition's name from its ID via the tracking store."""
    try:
        from mlflow.server.handlers import _get_tracking_store

        model_def = _get_tracking_store().get_gateway_model_definition(model_definition_id=model_definition_id)
        return model_def.name
    except Exception:
        _logger.debug(f"Could not resolve gateway model definition name from id '{model_definition_id}'")
        return None


def _get_gateway_endpoint_name() -> str | None:
    """Extract gateway endpoint name from the request.

    Tries ``name`` first, then falls back to resolving ``endpoint_id``
    via the tracking store.
    """
    try:
        return get_request_param("name")
    except Exception:
        pass
    try:
        endpoint_id = get_request_param("endpoint_id")
        return _resolve_endpoint_name_from_id(endpoint_id)
    except Exception:
        return None


def _get_gateway_endpoint_name_for_update() -> str | None:
    """Extract the *current* gateway endpoint name for update requests.

    Unlike ``_get_gateway_endpoint_name``, this deliberately skips the
    ``name`` request parameter because in ``UpdateGatewayEndpoint`` the
    ``name`` field holds the *new* name.  We resolve the current name
    via ``endpoint_id`` instead.
    """
    try:
        endpoint_id = get_request_param("endpoint_id")
        return _resolve_endpoint_name_from_id(endpoint_id)
    except Exception:
        return None


def _get_gateway_secret_name() -> str | None:
    """Extract gateway secret name from the request.

    Tries ``secret_name`` first, then falls back to resolving ``secret_id``
    via the tracking store.
    """
    try:
        return get_request_param("secret_name")
    except Exception:
        pass
    try:
        secret_id = get_request_param("secret_id")
        return _resolve_secret_name_from_id(secret_id)
    except Exception:
        return None


def _get_gateway_model_definition_name() -> str | None:
    """Extract gateway model definition name from the request.

    Tries ``name`` first, then falls back to resolving ``model_definition_id``
    via the tracking store.
    """
    try:
        return get_request_param("name")
    except Exception:
        pass
    try:
        model_definition_id = get_request_param("model_definition_id")
        return _resolve_model_definition_name_from_id(model_definition_id)
    except Exception:
        return None



from typing import Callable, List

from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import INVALID_STATE, RESOURCE_ALREADY_EXISTS, RESOURCE_DOES_NOT_EXIST
from sqlalchemy.exc import IntegrityError, MultipleResultsFound, NoResultFound
from sqlalchemy.orm import Session

from mlflow_oidc_auth.db.models import SqlGatewayEndpointPermission, SqlUser
from mlflow_oidc_auth.entities import GatewayEndpointPermission
from mlflow_oidc_auth.permissions import _validate_permission
from mlflow_oidc_auth.repository.utils import get_user


class GatewayEndpointPermissionRepository:
    def __init__(self, session_maker):
        self._Session: Callable[[], Session] = session_maker

    def _get_permission(self, session: Session, endpoint_id: str, username: str) -> SqlGatewayEndpointPermission:
        try:
            return (
                session.query(SqlGatewayEndpointPermission)
                .join(SqlUser, SqlGatewayEndpointPermission.user_id == SqlUser.id)
                .filter(
                    SqlGatewayEndpointPermission.endpoint_id == endpoint_id,
                    SqlUser.username == username,
                )
                .one()
            )
        except NoResultFound as e:
            raise MlflowException(
                f"No gateway endpoint permission for endpoint={endpoint_id}, user={username}",
                RESOURCE_DOES_NOT_EXIST,
            ) from e
        except MultipleResultsFound as e:
            raise MlflowException(
                f"Multiple gateway endpoint perms for endpoint={endpoint_id}, user={username}",
                INVALID_STATE,
            ) from e

    def grant_permission(self, endpoint_id: str, username: str, permission: str) -> GatewayEndpointPermission:
        _validate_permission(permission)
        with self._Session() as session:
            try:
                user = get_user(session, username)
                perm = SqlGatewayEndpointPermission(
                    endpoint_id=endpoint_id,
                    user_id=user.id,
                    permission=permission,
                )
                session.add(perm)
                session.flush()
                return perm.to_mlflow_entity()
            except IntegrityError as e:
                raise MlflowException(
                    f"Gateway endpoint permission already exists ({endpoint_id}, {username}): {e}",
                    RESOURCE_ALREADY_EXISTS,
                ) from e

    def get_permission(self, endpoint_id: str, username: str) -> GatewayEndpointPermission:
        with self._Session() as session:
            perm = self._get_permission(session, endpoint_id, username)
            return perm.to_mlflow_entity()

    def list_permissions_for_user(self, username: str) -> List[GatewayEndpointPermission]:
        with self._Session() as session:
            user = get_user(session, username)
            rows = session.query(SqlGatewayEndpointPermission).filter(SqlGatewayEndpointPermission.user_id == user.id).all()
            return [r.to_mlflow_entity() for r in rows]

    def update_permission(self, endpoint_id: str, username: str, permission: str) -> GatewayEndpointPermission:
        _validate_permission(permission)
        with self._Session() as session:
            perm = self._get_permission(session, endpoint_id, username)
            perm.permission = permission
            session.flush()
            return perm.to_mlflow_entity()

    def revoke_permission(self, endpoint_id: str, username: str) -> None:
        with self._Session() as session:
            perm = self._get_permission(session, endpoint_id, username)
            session.delete(perm)
            session.flush()

    def rename(self, old_name: str, new_name: str) -> None:
        """Update all user-level permissions from *old_name* to *new_name*."""
        with self._Session() as session:
            perms = session.query(SqlGatewayEndpointPermission).filter(SqlGatewayEndpointPermission.endpoint_id == old_name).all()
            for perm in perms:
                perm.endpoint_id = new_name
            session.flush()

    def wipe(self, endpoint_id: str) -> None:
        """Delete all user-level permissions for a gateway endpoint."""
        with self._Session() as session:
            perms = session.query(SqlGatewayEndpointPermission).filter(SqlGatewayEndpointPermission.endpoint_id == endpoint_id).all()
            for p in perms:
                session.delete(p)
            session.flush()



"""Pydantic models for scorer permission APIs."""

from typing import Optional

from pydantic import BaseModel, Field


class ScorerPermission(BaseModel):
    """Model for creating or updating a scorer permission."""

    permission: str = Field(..., description="Permission level for the scorer")


class ScorerRegexCreate(BaseModel):
    """Model for creating or updating a regex-based scorer permission."""

    regex: str = Field(..., description="Regex pattern to match scorer names")
    priority: int = Field(..., description="Priority of the permission rule")
    permission: str = Field(..., description="Permission level for matching scorers")


class ScorerPermissionRequest(BaseModel):
    """Request payload for scorer permission CRUD endpoints.

    MLflow v3 scorer permission routes expect these fields.
    """

    experiment_id: str = Field(..., description="Experiment ID owning the scorer")
    scorer_name: str = Field(..., description="Scorer name")
    username: str = Field(..., description="Target username")
    permission: str = Field(..., description="Permission name (e.g. READ/UPDATE/DELETE/MANAGE)")


class ScorerSummary(BaseModel):
    """Lightweight representation of a scorer entry for listing endpoints."""

    experiment_id: str = Field(..., description="Experiment identifier that owns the scorer")
    name: str = Field(..., description="Scorer name")
    version: int = Field(..., description="Scorer version number")
    creation_time: int = Field(..., description="Creation timestamp in milliseconds since epoch")
    scorer_id: Optional[str] = Field(None, description="Unique scorer identifier if present")
